import recurly
from recurly import Account, Transaction, ValidationError
from recurly.errors import UnexpectedStatusError, UnexpectedClientError, UnexpectedServerError
from recurlytests import RecurlyTest

class RecurlyExceptionTests(RecurlyTest):
    def test_error_printable(self):
        """ Make sure __str__/__unicode__ works correctly in Python 2/3"""
        str(recurly.UnauthorizedError('recurly.API_KEY not set'))

    def test_validationerror_printable(self):
        """ Make sure __str__/__unicode__ works correctly in Python 2/3"""
        error = recurly.ValidationError.Suberror('field', 'symbol', 'message')
        suberrors = dict()
        suberrors['field'] = error
        validation_error = recurly.ValidationError('')
        validation_error.__dict__['errors'] = suberrors
        str(validation_error)

    def test_transaction_error_property(self):
        """ Test ValidationError class 'transaction_error' property"""
        transaction = Transaction(
            amount_in_cents=1000,
            currency='USD',
            account=Account(
                account_code='transactionmock'
            )
        )

        # Mock 'save transaction' request to throw declined
        # transaction validation error
        with self.mock_request('transaction/declined-transaction.xml'):
            try:
                transaction.save()
            except ValidationError as e:
                error = e

        transaction_error = error.transaction_error

        self.assertEqual(transaction_error.error_code, 'insufficient_funds')
        self.assertEqual(transaction_error.error_category, 'soft')
        self.assertEqual(transaction_error.customer_message, "The transaction was declined due to insufficient funds in your account. Please use a different card or contact your bank.")
        self.assertEqual(transaction_error.merchant_message, "The card has insufficient funds to cover the cost of the transaction.")
        self.assertEqual(transaction_error.gateway_error_code, "123")

    def test_transaction_error_code_property(self):
        """ Test ValidationError class 'transaction_error_code' property"""
        transaction = Transaction(
            amount_in_cents=1000,
            currency='USD',
            account=Account(
                account_code='transactionmock'
            )
        )

        # Mock 'save transaction' request to throw declined
        # transaction validation error
        with self.mock_request('transaction/declined-transaction.xml'):
            try:
                transaction.save()
            except ValidationError as e:
                error = e

        self.assertEqual(error.transaction_error_code, 'insufficient_funds')

    def test_unexpected_errors_thrown(self):
        """ Test UnexpectedClientError class """
        transaction = Transaction(
            amount_in_cents=1000,
            currency='USD',
            account=Account(
                account_code='transactionmock'
            )
        )

        # Mock 'save transaction' request to throw unexpected client error
        with self.mock_request('transaction/error-teapot.xml'):
            try:
                transaction.save()
            except UnexpectedStatusError as e:
                error = e

        self.assertIsInstance(error, UnexpectedClientError)

        # Mock 'save transaction' request to throw another unexpected client error
        with self.mock_request('transaction/error-client.xml'):
            try:
                transaction.save()
            except UnexpectedStatusError as e:
                error = e

        self.assertIsInstance(error, UnexpectedClientError)

        # Mock 'save transaction' request to throw unexpected server error
        with self.mock_request('transaction/error-server.xml'):
            try:
                transaction.save()
            except UnexpectedStatusError as e:
                error = e

        self.assertIsInstance(error, UnexpectedServerError)



import collections
from datetime import datetime

import six
from six import StringIO
from six.moves.urllib.parse import urljoin

import recurly
from recurly import Account, AddOn, Address, Adjustment, BillingInfo, Coupon, Plan, Redemption, Subscription, \
    SubscriptionAddOn, Transaction, MeasuredUnit, Usage, GiftCard, Delivery, ShippingAddress, AccountAcquisition, \
    Purchase, Invoice, InvoiceCollection, CreditPayment, CustomField, ExportDate, ExportDateFile
from recurly import Money, NotFoundError, ValidationError, BadRequestError, PageError
from recurly import recurly_logging as logging
from recurlytests import RecurlyTest

recurly.SUBDOMAIN = 'api'


class TestResources(RecurlyTest):

    def test_authentication(self):
        recurly.API_KEY = None

        account_code = 'test%s' % self.test_id
        try:
            Account.get(account_code)
        except recurly.UnauthorizedError as exc:
            pass
        else:
            self.fail("Updating account with invalid email address did not raise a ValidationError")

    def test_config_string_types(self):
        recurly.API_KEY = six.u('\xe4 unicode string')

        account_code = 'test%s' % self.test_id
        try:
            Account.get(account_code)
        except recurly.ConfigurationError as exc:
            pass
        else:
            self.fail("Updating account with invalid email address did not raise a ValidationError")

    def test_cached_response_headers(self):
        account_code = 'test%s' % self.test_id
        with self.mock_request('account/exists-with-rate-limit-headers.xml'):
            account = Account.get(account_code)

        self.assertEqual(recurly.cached_rate_limits['limit'], 2000)
        self.assertEqual(recurly.cached_rate_limits['remaining'], 1992)
        self.assertEqual(recurly.cached_rate_limits['resets_at'], datetime(2017, 2, 2, 19, 46))
        self.assertIsInstance(recurly.cached_rate_limits['cached_at'], datetime)

    def test_credit_payment(self):
        with self.mock_request('credit-payment/show.xml'):
            payment = CreditPayment.get('43c29728b3482fa8727edb4cefa7c774')
            self.assertIsInstance(payment, CreditPayment)
            self.assertEquals(payment.uuid, '43c29728b3482fa8727edb4cefa7c774')
            self.assertEquals(payment.amount_in_cents, 12)
            self.assertEquals(payment.currency, 'USD')
            self.assertEquals(payment.action, 'payment')

    def test_purchase(self):
        account_code = 'test%s' % self.test_id
        purchase = Purchase(
            currency = 'USD',
            gateway_code = 'aBcD1234',
            collection_method = 'manual',
            account = Account(
                account_code = account_code,
                shipping_addresses = [
                    ShippingAddress(
                        first_name = 'Verena',
                        last_name = 'Example',
                        address1 = '123 Main St',
                        city = 'New Orleans',
                        state = 'LA',
                        zip = '70114',
                        country = 'US',
                        nickname = 'Work'
                    )
                ],
                billing_info = BillingInfo(
                    first_name = 'Verena',
                    last_name = 'Example',
                    number = '4111-1111-1111-1111',
                    verification_value = '123',
                    month = 11,
                    year = 2020,
                    address1 = '123 Main St',
                    city = 'New Orleans',
                    state = 'LA',
                    zip = '70114',
                    country = 'US',
                )
            ),
            subscriptions = [
                recurly.Subscription(plan_code = 'gold')
            ],
            adjustments = [
                recurly.Adjustment(unit_amount_in_cents=1000, description='Item 1',
                                   quantity=1),
                recurly.Adjustment(unit_amount_in_cents=2000, description='Item 2',
                                   quantity=2),
            ]
        )
        with self.mock_request('purchase/invoiced.xml'):
            collection = purchase.invoice()
            self.assertIsInstance(collection, InvoiceCollection)
            self.assertIsInstance(collection.charge_invoice, Invoice)
            self.assertIsInstance(collection.credit_invoices, list)
            self.assertIsInstance(collection.credit_invoices[0], Invoice)
            self.assertIsInstance(collection.charge_invoice.line_items[0].shipping_address,
                                  ShippingAddress)
        with self.mock_request('purchase/previewed.xml'):
            collection = purchase.preview()
            self.assertIsInstance(collection, InvoiceCollection)
            self.assertIsInstance(collection.charge_invoice, Invoice)
        with self.mock_request('purchase/authorized.xml'):
            purchase.account.email = 'benjamin.dumonde@example.com'
            purchase.account.billing_info.external_hpp_type = 'adyen'
            collection = purchase.authorize()
            self.assertIsInstance(collection, InvoiceCollection)
            self.assertIsInstance(collection.charge_invoice, Invoice)
        with self.mock_request('purchase/pending.xml'):
            purchase.account.email = 'benjamin.dumonde@example.com'
            purchase.account.billing_info.external_hpp_type = 'adyen'
            collection = purchase.pending()
            self.assertIsInstance(collection, InvoiceCollection)
            self.assertIsInstance(collection.charge_invoice, Invoice)

    def test_account(self):
        account_code = 'test%s' % self.test_id
        with self.mock_request('account/does-not-exist.xml'):
            self.assertRaises(NotFoundError, Account.get, account_code)

        account = Account(account_code=account_code)
        account.vat_number = '444444-UK'
        account.preferred_locale = 'en-US'
        with self.mock_request('account/created.xml'):
            account.save()
        self.assertEqual(account._url, urljoin(recurly.base_uri(), 'accounts/%s' % account_code))
        self.assertEqual(account.vat_number, '444444-UK')
        self.assertEqual(account.vat_location_enabled, True)
        self.assertEqual(account.cc_emails,
                'test1@example.com,test2@example.com')
        self.assertEqual(account.preferred_locale, 'en-US')

        with self.mock_request('account/list-active.xml'):
            active = Account.all_active()
        self.assertTrue(len(active) >= 1)
        self.assertEqual(active[0].account_code, account_code)

        with self.mock_request('account/exists.xml'):
            same_account = Account.get(account_code)
        self.assertTrue(isinstance(same_account, Account))
        self.assertTrue(same_account is not account)
        self.assertEqual(same_account.account_code, account_code)
        self.assertTrue(same_account.first_name is None)
        self.assertTrue(same_account.entity_use_code == 'I')
        self.assertEqual(same_account._url, urljoin(recurly.base_uri(), 'accounts/%s' % account_code))

        with self.mock_request('account-balance/exists.xml'):
            account_balance = same_account.account_balance()

        self.assertTrue(account_balance.past_due)
        balance = account_balance.balance_in_cents
        self.assertTrue(balance['USD'] == 2910)
        self.assertTrue(balance['EUR'] == -520)

        account.username = 'shmohawk58'
        account.email = 'larry.david'
        account.first_name = six.u('L\xe4rry')
        account.last_name = 'David'
        account.company_name = 'Home Box Office'
        account.accept_language = 'en-US'
        with self.mock_request('account/update-bad-email.xml'):
            try:
                account.save()
            except ValidationError as exc:
                self.assertTrue(isinstance(exc.errors, collections.Mapping))
                self.assertTrue('account.email' in exc.errors)
                suberror = exc.errors['account.email']
                self.assertEqual(suberror.symbol, 'invalid_email')
                self.assertTrue(suberror.message)
                self.assertEqual(suberror.message, suberror.message)
            else:
                self.fail("Updating account with invalid email address did not raise a ValidationError")

        account.email = 'larry.david@example.com'
        with self.mock_request('account/updated.xml'):
            account.save()

        with self.mock_request('account/deleted.xml'):
            account.delete()

        with self.mock_request('account/list-closed.xml'):
            closed = Account.all_closed()
        self.assertTrue(len(closed) >= 1)
        self.assertEqual(closed[0].account_code, account_code)

        with self.mock_request('account/list-active-when-closed.xml'):
            active = Account.all_active()
        self.assertTrue(len(active) < 1 or active[0].account_code != account_code)

        # Make sure we can reopen a closed account.
        with self.mock_request('account/reopened.xml'):
            account.reopen()
        try:
            with self.mock_request('account/list-active.xml'):
                active = Account.all_active()
            self.assertTrue(len(active) >= 1)
            self.assertEqual(active[0].account_code, account_code)
        finally:
            with self.mock_request('account/deleted.xml'):
                account.delete()

        # Make sure numeric account codes work.
        if self.test_id == 'mock':
            numeric_test_id = 58
        else:
            numeric_test_id = int(self.test_id)

        account = Account(account_code=numeric_test_id)
        with self.mock_request('account/numeric-created.xml'):
            account.save()
        try:
            self.assertEqual(account._url, urljoin(recurly.base_uri(), 'accounts/%d' % numeric_test_id))
        finally:
            with self.mock_request('account/numeric-deleted.xml'):
                account.delete()

        """Get taxed account"""
        with self.mock_request('account/show-taxed.xml'):
            account = Account.get(account_code)
            self.assertTrue(account.tax_exempt)
            self.assertEqual(account.exemption_certificate, 'Some Certificate')

    def test_account_addresses(self):
        account_code = 'test%s' % self.test_id
        """Create an account with an account level address"""
        account = Account(account_code=account_code)
        account.address.address1 = '123 Main St'
        account.address.city = 'San Francisco'
        account.address.zip = '94105'
        account.address.state = 'CA'
        account.address.country = 'US'
        account.address.phone = '8015559876'

        with self.mock_request('account/created-with-address.xml'):
            account.save()
        self.assertEqual(account.address.address1, '123 Main St')
        self.assertEqual(account.address.city, 'San Francisco')
        self.assertEqual(account.address.zip, '94105')
        self.assertEqual(account.address.state, 'CA')
        self.assertEqual(account.address.country, 'US')
        self.assertEqual(account.address.phone, '8015559876')

        """Create an account with an account shipping address"""
        account = Account(account_code=account_code)
        shipping_address = ShippingAddress()
        shipping_address.address1 = '123 Main St'
        shipping_address.city = 'San Francisco'
        shipping_address.zip = '94105'
        shipping_address.state = 'CA'
        shipping_address.country = 'US'
        shipping_address.phone = '8015559876'
        shipping_address.nickname = 'Work'

        account.shipping_addresses = [shipping_address]

        with self.mock_request('account/created-with-shipping-address.xml'):
            account.save()

        shipping_address = ShippingAddress()
        shipping_address.address1 = '123 Dolores St'
        shipping_address.city = 'San Francisco'
        shipping_address.zip = '94105'
        shipping_address.state = 'CA'
        shipping_address.country = 'US'
        shipping_address.phone = '8015559876'
        shipping_address.nickname = 'Home'

        with self.mock_request('shipping_addresses/created-on-existing-account.xml'):
            shipping_address = account.create_shipping_address(shipping_address)

    def test_account_custom_fields(self):
        account_code = 'test%s' % self.test_id
        """Create an account with a custom field"""
        account = Account(
            account_code=account_code,
            custom_fields=[
                CustomField(name="field_1", value="my field value")
            ]
        )
        with self.mock_request('account/created-with-custom-fields.xml'):
            account.save()

        self.assertEquals(account.custom_fields[0].name, 'field_1')
        self.assertEquals(account.custom_fields[0].value, 'my field value')

        """Update custom fields on an account"""
        with self.mock_request('account/exists-custom-fields.xml'):
            existing_account = Account.get(account_code)
        fields = existing_account.custom_fields
        fields[1].value = "new value2"
        existing_account.custom_fields = fields
        with self.mock_request('account/updated-custom-fields.xml'):
            existing_account.save()

        self.assertEquals(existing_account.custom_fields[0].name, 'field1')
        self.assertEquals(existing_account.custom_fields[0].value, 'original value1')
        self.assertEquals(existing_account.custom_fields[1].name, 'field2')
        self.assertEquals(existing_account.custom_fields[1].value, 'new value2')

    def test_account_hierarchy(self):
        account_code = 'test%s' % self.test_id
        """Create an account with a parent"""
        account = Account(
            account_code=account_code,
            parent_account_code="parent-account"
        )
        with self.mock_request('account/created-with-parent.xml'):
            account.save()

        """Get Parent account"""
        with self.mock_request('account/exists-with-child-accounts.xml'):
            parent = account.parent_account()

        self.assertEquals(parent.account_code, 'parent-account')

        """Get Child accounts"""
        with self.mock_request('account/child-accounts.xml'):
            for acct in parent.child_accounts():
                self.assertEquals(acct.account_code, 'test%s' % self.test_id)

    def test_account_acquisition(self):
        account_code = 'test%s' % self.test_id

        """Create an account with an account acquisition"""
        account = Account(account_code=account_code)
        acquisition = AccountAcquisition()
        acquisition.cost_in_cents = 199
        acquisition.currency = 'USD'
        acquisition.channel = 'blog'
        acquisition.subchannel = 'Whitepaper Blog Post'
        acquisition.campaign = 'mailchimp67a904de95.0914d8f4b4'
        account.account_acquisition = acquisition

        with self.mock_request('account/created-with-account-acquisition.xml'):
            account.save()

        """Get the acquisition from the account"""
        with self.mock_request('account-acquisition/exists.xml'):
            acquisition = account.account_acquisition()

        self.assertEquals(acquisition.cost_in_cents, 199)
        self.assertEquals(acquisition.currency, 'USD')
        self.assertEquals(acquisition.channel, 'blog')
        self.assertEquals(acquisition.subchannel, 'Whitepaper Blog Post')
        self.assertEquals(acquisition.campaign, 'mailchimp67a904de95.0914d8f4b4')

        """Update the acquisition"""
        acquisition.cost_in_cents = 200
        acquisition.currency = 'EUR'
        acquisition.channel = 'social_media'
        acquisition.subchannel = 'Facebook Post'
        acquisition.campaign = 'hubspot123456'

        with self.mock_request('account-acquisition/updated.xml'):
            acquisition.save()

        self.assertEquals(acquisition.cost_in_cents, 200)
        self.assertEquals(acquisition.currency, 'EUR')
        self.assertEquals(acquisition.channel, 'social_media')
        self.assertEquals(acquisition.subchannel, 'Facebook Post')
        self.assertEquals(acquisition.campaign, 'hubspot123456')

        with self.mock_request('account-acquisition/deleted.xml'):
            acquisition.delete()

    def test_add_on(self):
        plan_code = 'plan%s' % self.test_id
        add_on_code = 'addon%s' % self.test_id

        plan = Plan(
            plan_code=plan_code,
            name='Mock Plan',
            setup_fee_in_cents=Money(0),
            unit_amount_in_cents=Money(1000),
        )
        with self.mock_request('add-on/plan-created.xml'):
            plan.save()

        try:
            # a usage based add on
            add_on = AddOn(
                add_on_code=add_on_code,
                name='Mock Add-On',
                add_on_type="usage",
                usage_type="price",
                measured_unit_id=123456,
            )
            exc = None
            with self.mock_request('add-on/need-amount.xml'):
                try:
                    plan.create_add_on(add_on)
                except ValidationError as _exc:
                    exc = _exc
                else:
                    self.fail("Creating a plan add-on without an amount did not raise a ValidationError")
            error = exc.errors['add_on.unit_amount_in_cents']
            self.assertEqual(error.symbol, 'blank')

            add_on.unit_amount_in_cents = Money(40)

            with self.mock_request('add-on/created.xml'):
                plan.create_add_on(add_on)
            self.assertEqual(add_on.add_on_code, add_on_code)
            self.assertEqual(add_on.name, 'Mock Add-On')

            try:

                with self.mock_request('add-on/exists.xml'):
                    same_add_on = plan.get_add_on(add_on_code)
                self.assertEqual(same_add_on.add_on_code, add_on_code)
                self.assertEqual(same_add_on.name, 'Mock Add-On')
                self.assertEqual(same_add_on.unit_amount_in_cents['USD'], 40)

            finally:
                with self.mock_request('add-on/deleted.xml'):
                    add_on.delete()
        finally:
            with self.mock_request('add-on/plan-deleted.xml'):
                plan.delete()

    def test_billing_info(self):
        logging.basicConfig(level=logging.DEBUG)  # make sure it's init'ed
        logger = logging.getLogger('recurly.http.request')
        logger.setLevel(logging.DEBUG)

        log_content = StringIO()
        log_handler = logging.StreamHandler(log_content)
        logger.addHandler(log_handler)

        account = Account(account_code='binfo%s' % self.test_id)
        with self.mock_request('billing-info/account-created.xml'):
            account.save()

        logger.removeHandler(log_handler)
        self.assertTrue('<account' in log_content.getvalue())

        try:

            # Billing info link won't be present at all yet.
            self.assertRaises(AttributeError, getattr, account, 'billing_info')

            log_content = StringIO()
            log_handler = logging.StreamHandler(log_content)
            logger.addHandler(log_handler)

            binfo = BillingInfo(
                first_name='Verena',
                last_name='Example',
                address1='123 Main St',
                city=six.u('San Jos\xe9'),
                state='CA',
                zip='94105',
                country='US',
                type='credit_card',
                number='4111 1111 1111 1111',
                verification_value='7777',
                year='2015',
                month='12',
                gateway_token='gatewaytoken123',
                gateway_code='gatewaycode123',
            )
            with self.mock_request('billing-info/created.xml'):
                account.update_billing_info(binfo)

            self.assertEqual(binfo.gateway_token, 'gatewaytoken123')
            self.assertEqual(binfo.gateway_code, 'gatewaycode123')

            logger.removeHandler(log_handler)
            log_content = log_content.getvalue()
            self.assertTrue('<billing_info' in log_content)
            # See if we redacted our sensitive fields properly.
            self.assertTrue('4111' not in log_content)
            self.assertTrue('7777' not in log_content)

            with self.mock_request('billing-info/account-exists.xml'):
                same_account = Account.get('binfo%s' % self.test_id)
            with self.mock_request('billing-info/exists.xml'):
                same_binfo = same_account.billing_info
            self.assertEqual(same_binfo.first_name, 'Verena')
            self.assertEqual(same_binfo.city, six.u('San Jos\xe9'))

            with self.mock_request('billing-info/deleted.xml'):
                binfo.delete()
        finally:
            with self.mock_request('billing-info/account-deleted.xml'):
                account.delete()

        # Credit Card
        log_content = StringIO()
        log_handler = logging.StreamHandler(log_content)
        logger.addHandler(log_handler)

        account = Account(account_code='binfo-%s-2' % self.test_id)
        account.billing_info = BillingInfo(
            first_name='Verena',
            last_name='Example',
            address1='123 Main St',
            city=six.u('San Jos\xe9'),
            state='CA',
            zip='94105',
            country='US',
            type='credit_card',
            number='4111 1111 1111 1111',
            verification_value='7777',
            year='2015',
            month='12',
        )
        with self.mock_request('billing-info/account-embed-created.xml'):
            account.save()

        try:
            logger.removeHandler(log_handler)
            log_content = log_content.getvalue()
            self.assertTrue('<account' in log_content)
            self.assertTrue('<billing_info' in log_content)
            self.assertTrue('4111' not in log_content)
            self.assertTrue('7777' not in log_content)

            with self.mock_request('billing-info/account-embed-exists.xml'):
                same_account = Account.get('binfo-%s-2' % self.test_id)
            with self.mock_request('billing-info/embedded-exists.xml'):
                binfo = same_account.billing_info
            self.assertEqual(binfo.first_name, 'Verena')
        finally:
            with self.mock_request('billing-info/account-embed-deleted.xml'):
                account.delete()

        # Token
        log_content = StringIO()
        log_handler = logging.StreamHandler(log_content)
        logger.addHandler(log_handler)

        account = Account(account_code='binfo-%s-3' % self.test_id)
        account.billing_info = BillingInfo(token_id = 'abc123')
        with self.mock_request('billing-info/account-embed-token.xml'):
            account.save()

        logger.removeHandler(log_handler)
        log_content = log_content.getvalue()
        self.assertTrue('<billing_info' in log_content)
        self.assertTrue('<token_id' in log_content)

    def test_charge(self):
        account = Account(account_code='charge%s' % self.test_id)
        with self.mock_request('adjustment/account-created.xml'):
            account.save()

        try:
            with self.mock_request('adjustment/account-has-no-charges.xml'):
                charges = account.adjustments()
            self.assertEqual(charges, [])

            charge = Adjustment(unit_amount_in_cents=1000, currency='USD', description='test charge', type='charge')
            with self.mock_request('adjustment/charged.xml'):
                account.charge(charge)

            with self.mock_request('adjustment/account-has-adjustments.xml'):
                charges = account.adjustments()
            self.assertEqual(len(charges), 1)
            same_charge = charges[0]
            self.assertEqual(same_charge.unit_amount_in_cents, 1000)
            self.assertEqual(same_charge.tax_in_cents, 5000)
            self.assertEqual(same_charge.tax_type, 'usst')
            self.assertEqual(same_charge.tax_rate, 0.0875)
            self.assertEqual(same_charge.tax_region, 'CA')
            self.assertEqual(same_charge.currency, 'USD')
            self.assertEqual(same_charge.description, 'test charge')
            self.assertEqual(same_charge.type, 'charge')

            tax_details = same_charge.tax_details
            state, county = tax_details

            self.assertEqual(len(tax_details), 2)
            self.assertEqual(state.name, 'california')
            self.assertEqual(state.type, 'state')
            self.assertEqual(state.tax_rate, 0.065)
            self.assertEqual(state.tax_in_cents, 3000)

            self.assertEqual(county.name, 'san francisco')
            self.assertEqual(county.type, 'county')
            self.assertEqual(county.tax_rate, 0.02)
            self.assertEqual(county.tax_in_cents, 2000)

            with self.mock_request('adjustment/account-has-charges.xml'):
                charges = account.adjustments(type='charge')
            self.assertEqual(len(charges), 1)

            with self.mock_request('adjustment/account-has-no-credits.xml'):
                credits = account.adjustments(type='credit')
            self.assertEqual(len(credits), 0)

        finally:
            with self.mock_request('adjustment/account-deleted.xml'):
                account.delete()

        """Test taxed adjustments"""
        with self.mock_request('adjustment/show-taxed.xml'):
            charge = account.adjustments()[0]
            self.assertFalse(charge.tax_exempt)

        """Test original adjustment"""
        with self.mock_request('adjustment/original-adjustment.xml'):
            charge = Adjustment.get('2c06b94abe047189b225d94dd0adb71f')

            with self.mock_request('adjustment/original-adjustment-lookup.xml'):
                original_charge = charge.original_adjustment()

            self.assertEqual(original_charge.total_in_cents, -charge.total_in_cents)

        """Test original adjustment lookup by UUID"""
        with self.mock_request('adjustment/original-adjustment-uuid.xml'):
            charge = Adjustment.get('2c06b94abe047189b225d94dd0adb71f')

            with self.mock_request('adjustment/original-adjustment-lookup.xml'):
                original_charge = charge.original_adjustment()

            self.assertEqual(original_charge.total_in_cents, -charge.total_in_cents)

    def test_coupon(self):
        # Check that a coupon may not exist.
        coupon_code = 'coupon%s' % self.test_id
        with self.mock_request('coupon/does-not-exist.xml'):
            self.assertRaises(NotFoundError, Coupon.get, coupon_code)

        # Create a coupon?
        coupon = Coupon(
            coupon_code=coupon_code,
            name='Nice Coupon',
            discount_in_cents=Money(1000),
            hosted_description='Nice Description',
            invoice_description='Invoice description'
        )
        with self.mock_request('coupon/created.xml'):
            coupon.save()
        self.assertTrue(coupon._url)

        try:

            with self.mock_request('coupon/exists.xml'):
                same_coupon = Coupon.get(coupon_code)
            self.assertEqual(same_coupon.coupon_code, coupon_code)
            self.assertEqual(same_coupon.name, 'Nice Coupon')
            self.assertEqual(same_coupon.invoice_description, 'Invoice description')
            discount = same_coupon.discount_in_cents
            self.assertEqual(discount['USD'], 1000)
            self.assertTrue('USD' in discount)
            self.assertIsNotNone(same_coupon.hosted_description)

            account_code = 'coupon%s' % self.test_id
            account = Account(account_code=account_code)
            with self.mock_request('coupon/account-created.xml'):
                account.save()

            coupon.name = 'New Name'
            coupon.invoice_description = 'New Description'
            coupon.hosted_description = 'New Description'

            with self.mock_request('coupon/updated.xml'):
                coupon.save()

            self.assertEqual(coupon.name, 'New Name')
            self.assertEqual(coupon.invoice_description, 'New Description')
            self.assertEqual(coupon.hosted_description, 'New Description')

            coupon.name = 'New Name Restore'
            coupon.invoice_description = 'New Description Restore'
            coupon.hosted_description = 'New Description Restore'

            with self.mock_request('coupon/restored.xml'):
                coupon.restore()

            self.assertEqual(coupon.name, 'New Name Restore')
            self.assertEqual(coupon.invoice_description, 'New Description Restore')
            self.assertEqual(coupon.hosted_description, 'New Description Restore')

            try:

                redemption = Redemption(
                    account_code=account_code,
                    currency='USD',
                )
                with self.mock_request('coupon/redeemed.xml'):
                    real_redemption = coupon.redeem(redemption)
                self.assertTrue(isinstance(real_redemption, Redemption))
                self.assertEqual(real_redemption.currency, 'USD')

                with self.mock_request('coupon/account-with-redemption.xml'):
                    account = Account.get(account_code)
                with self.mock_request('coupon/redemption-exists.xml'):
                    same_redemption = account.redemption()
                self.assertEqual(same_redemption._url, real_redemption._url)

                with self.mock_request('coupon/unredeemed.xml'):
                    real_redemption.delete()

            finally:
                with self.mock_request('coupon/account-deleted.xml'):
                    account.delete()

            plan = Plan(
                plan_code='basicplan',
                name='Basic Plan',
                setup_fee_in_cents=Money(0),
                unit_amount_in_cents=Money(1000),
            )
            with self.mock_request('coupon/plan-created.xml'):
                plan.save()

            try:

                account_code_2 = 'coupon-%s-2' % self.test_id
                sub = Subscription(
                    plan_code='basicplan',
                    coupon_code='coupon%s' % self.test_id,
                    currency='USD',
                    account=Account(
                        account_code=account_code_2,
                        billing_info=BillingInfo(
                            first_name='Verena',
                            last_name='Example',
                            number='4111 1111 1111 1111',
                            address1='123 Main St',
                            city='San Francisco',
                            state='CA',
                            zip='94105',
                            country='US',
                            verification_value='7777',
                            year='2015',
                            month='12',
                        ),
                    ),
                )
                with self.mock_request('coupon/subscribed.xml'):
                    sub.save()

                with self.mock_request('coupon/second-account-exists.xml'):
                    account_2 = Account.get(account_code_2)

                try:

                    with self.mock_request('coupon/second-account-redemption.xml'):
                        redemption_2 = account_2.redemption()
                    self.assertTrue(isinstance(redemption_2, Redemption))
                    self.assertEqual(redemption_2.currency, 'USD')
                    with self.mock_request('coupon/exists.xml'):
                        same_coupon = redemption_2.coupon()
                    self.assertEqual(same_coupon.coupon_code, coupon_code)

                finally:
                    with self.mock_request('coupon/second-account-deleted.xml'):
                        account_2.delete()

                plan_coupon = Coupon(
                    coupon_code='plancoupon%s' % self.test_id,
                    name='Plan Coupon',
                    discount_in_cents=Money(1000),
                    invoice_description='Invoice description',
                    applies_to_all_plans=False,
                    plan_codes=('basicplan',),
                    applies_to_non_plan_charges=True,
                    redemption_resource='subscription',
                    max_redemptions_per_account=1,
                )
                with self.mock_request('coupon/plan-coupon-created.xml'):
                    plan_coupon.save()

                try:
                    self.assertTrue(plan_coupon._url)
                    self.assertFalse(plan_coupon.has_unlimited_redemptions_per_account())

                    coupon_plans = list(plan_coupon.plan_codes)
                    self.assertEqual(len(coupon_plans), 1)
                    self.assertEqual(coupon_plans[0], 'basicplan')
                finally:
                    with self.mock_request('coupon/plan-coupon-deleted.xml'):
                        plan_coupon.delete()

            finally:
                with self.mock_request('coupon/plan-deleted.xml'):
                    plan.delete()

        finally:
            with self.mock_request('coupon/deleted.xml'):
                coupon.delete()

    def test_invoice(self):
        account = Account(account_code='invoice%s' % self.test_id)
        with self.mock_request('invoice/account-created.xml'):
            account.save()

        try:
            with self.mock_request('invoice/account-has-no-invoices.xml'):
                invoices = account.invoices()
            self.assertEqual(invoices, [])

            with self.mock_request('invoice/error-no-charges.xml'):
                try:
                    account.invoice()
                except ValidationError as exc:
                    error = exc
                else:
                    self.fail("Invoicing an account with no charges did not raise a ValidationError")
            self.assertEqual(error.symbol, 'will_not_invoice')

            charge = Adjustment(unit_amount_in_cents=1000, currency='USD', description='test charge', type='charge')
            with self.mock_request('invoice/charged.xml'):
                account.charge(charge)

            with self.mock_request('invoice/invoiced.xml'):
                account.invoice()

            with self.mock_request('invoice/account-has-invoices.xml'):
                invoices = account.invoices()
            self.assertEqual(len(invoices), 1)
        finally:
            with self.mock_request('invoice/account-deleted.xml'):
                account.delete()

        """Test taxed invoice"""
        with self.mock_request('invoice/show-taxed.xml'):
            invoice = account.invoices()[0]
            self.assertEqual(invoice.tax_type, 'usst')

        """Test invoice with prefix"""
        with self.mock_request('invoice/show-with-prefix.xml'):
            invoice = account.invoices()[0]
            self.assertEqual(invoice.invoice_number, 1001)
            self.assertEqual(invoice.invoice_number_prefix, 'GB')
            self.assertEqual(invoice.invoice_number_with_prefix(), 'GB1001')

    def test_invoice_refund_amount(self):
        account = Account(account_code='invoice%s' % self.test_id)
        with self.mock_request('invoice/account-created.xml'):
            account.save()

        with self.mock_request('invoice/invoiced.xml'):
            invoice = account.invoice().charge_invoice

        with self.mock_request('invoice/refunded.xml'):
            options = {
                'refund_method': 'credit_first',
                'credit_customer_notes': 'Credit Customer Notes',
                'description': 'Description'
            }
            refund_invoice = invoice.refund_amount(1000, options)
        self.assertEqual(refund_invoice.subtotal_in_cents, -1000)

    def test_invoice_refund(self):
        account = Account(account_code='invoice%s' % self.test_id)
        with self.mock_request('invoice/account-created.xml'):
            account.save()

        with self.mock_request('invoice/invoiced-line-items.xml'):
            invoice = account.invoice().charge_invoice

        with self.mock_request('invoice/line-item-refunded.xml'):
            line_items = [{ 'adjustment': invoice.line_items[0], 'quantity': 1,
                'prorate': False }]
            options = {
                'refund_method': 'credit_first',
                'credit_customer_notes': 'Credit Customer Notes',
                'description': 'Description'
            }
            refund_invoice = invoice.refund(line_items, options)
        self.assertEqual(refund_invoice.subtotal_in_cents, -1000)

    def test_invoice_with_optionals(self):
        account = Account(account_code='invoice%s' % self.test_id)
        with self.mock_request('invoice/account-created.xml'):
            account.save()

        with self.mock_request('invoice/invoiced-with-optionals.xml'):
            collection = account.invoice(terms_and_conditions='Some Terms and Conditions',
                    customer_notes='Some Customer Notes', collection_method="manual",
                    net_terms=30)

        invoice = collection.charge_invoice

        self.assertIsInstance(collection, InvoiceCollection)
        self.assertIsInstance(invoice, Invoice)
        self.assertEqual(invoice.terms_and_conditions, 'Some Terms and Conditions')
        self.assertEqual(invoice.customer_notes, 'Some Customer Notes')
        self.assertEqual(invoice.collection_method, 'manual')
        self.assertEqual(invoice.net_terms, 30)

    def test_invoice_offline_payment(self):
        with self.mock_request('invoice/show-invoice.xml'):
            invoice = Invoice.get("6019")

        self.assertIsInstance(invoice, Invoice)

        with self.mock_request('invoice/offline-payment.xml'):
            transaction = Transaction(amount_in_cents=5000, description="Collected externally")
            transaction = invoice.enter_offline_payment(transaction)

        self.assertIsInstance(transaction, Transaction)

    def test_invoice_create(self):
        # Invoices should not be created with save method
        invoice = Invoice()
        self.assertRaises(BadRequestError, invoice.save)

    def test_invoice_update(self):
        with self.mock_request('invoice/show-invoice.xml'):
            invoice = Invoice.get("6019")

        self.assertIsInstance(invoice, Invoice)

        with self.mock_request('invoice/update-invoice.xml'):
            invoice.address = recurly.Address(
                first_name = 'Harry',
                last_name = 'Potter',
                company = 'Hogwarts',
                name_on_account = 'Albus Dumbledore',
                address1 = '4 Privet Drive',
                address2 = 'Little Whinging',
                city = 'Surrey',
                state = 'England',
                zip = 'YO8 9FX',
                country = 'Great Britain',
                phone = '781-452-4077'
            )
            invoice.po_number = '1234'
            invoice.terms_and_conditions = 'School staff is not responsible for items left at Hogwarts School of Witchcraft and Wizardry.'
            invoice.customer_notes = "It's levi-O-sa, not levio-SA!"
            invoice.vat_reverse_charge_notes = "can't be changed when invoice was not a reverse charge"
            invoice.net_terms = 1
            invoice.gateway_code = 'A new gateway code'
            invoice.save()

        self.assertEqual(invoice.address.first_name, 'Harry')
        self.assertEqual(invoice.address.last_name, 'Potter')
        self.assertEqual(invoice.address.name_on_account, 'Albus Dumbledore')
        self.assertEqual(invoice.address.company, 'Hogwarts')
        self.assertEqual(invoice.address.address1, '4 Privet Drive')
        self.assertEqual(invoice.address.address2, 'Little Whinging')
        self.assertEqual(invoice.address.city, 'Surrey')
        self.assertEqual(invoice.address.state, 'England')
        self.assertEqual(invoice.address.zip, 'YO8 9FX')
        self.assertEqual(invoice.address.country, 'Great Britain')
        self.assertEqual(invoice.address.phone, '781-452-4077')
        self.assertEqual(invoice.po_number, '1234')
        self.assertEqual(invoice.terms_and_conditions, 'School staff is not responsible for items left at Hogwarts School of Witchcraft and Wizardry.')
        self.assertEqual(invoice.customer_notes, "It's levi-O-sa, not levio-SA!")
        self.assertEqual(invoice.vat_reverse_charge_notes, "can't be changed when invoice was not a reverse charge")
        self.assertEqual(invoice.net_terms, 1)
        self.assertEqual(invoice.gateway_code, 'A new gateway code')

    def test_build_invoice(self):
        account = Account(account_code='invoice%s' % self.test_id)
        with self.mock_request('invoice/account-created.xml'):
            account.save()

        try:
            with self.mock_request('invoice/preview-error-no-charges.xml'):
                try:
                    account.build_invoice()
                except ValidationError as exc:
                    error = exc
                else:
                    self.fail("Invoicing an account with no charges did not raise a ValidationError")
            self.assertEqual(error.symbol, 'will_not_invoice')

            charge = Adjustment(unit_amount_in_cents=1000, currency='USD', description='test charge', type='charge')
            with self.mock_request('invoice/charged.xml'):
                account.charge(charge)

            with self.mock_request('invoice/preview-invoice.xml'):
                account.build_invoice()
        finally:
            with self.mock_request('invoice/account-deleted.xml'):
                account.delete()

    def test_count(self):
        try:
            with self.mock_request('pages/count.xml'):
                num_accounts = Account.count(begin_time='2017-05-01T10:30:01-06:00')

            self.assertTrue(num_accounts, 23)
        finally:
            pass

    def test_pages(self):
        account_code = 'pages-%s-%%d' % self.test_id
        all_test_accounts = list()

        try:
            for i in range(1, 8):
                account = Account(account_code=account_code % i)
                all_test_accounts.append(account)
                with self.mock_request('pages/account-%d-created.xml' % i):
                    account.save()
                    self.mock_sleep(1)

            with self.mock_request('pages/list.xml'):
                accounts = Account.all(per_page=4)
            self.assertTrue(isinstance(accounts[0], Account))
            self.assertRaises(IndexError, lambda: accounts[4])

            # Test errors, since the first page has no first page.
            self.assertRaises(PageError, lambda: accounts.first_page())
            # Make sure PageError is a ValueError.
            self.assertRaises(ValueError, lambda: accounts.first_page())

            with self.mock_request('pages/next-list.xml'):
                next_accounts = accounts.next_page()
            # We asked for all the accounts, which may include closed accounts
            # from previous tests or data, not just the three we created.
            self.assertTrue(isinstance(next_accounts[0], Account))
            self.assertRaises(IndexError, lambda: next_accounts[4])

            with self.mock_request('pages/list.xml'):  # should be just like the first
                first_accounts = next_accounts.first_page()
            self.assertTrue(isinstance(first_accounts[0], Account))

        finally:
            for i, account in enumerate(all_test_accounts, 1):
                with self.mock_request('pages/account-%d-deleted.xml' % i):
                    account.delete()

    def test_plan(self):
        plan_code = 'plan%s' % self.test_id
        with self.mock_request('plan/does-not-exist.xml'):
            self.assertRaises(NotFoundError, Plan.get, plan_code)

        plan = Plan(
            plan_code=plan_code,
            name='Mock Plan',
            setup_fee_in_cents=Money(0),
            unit_amount_in_cents=Money(1000),
            total_billing_cycles=10
        )
        with self.mock_request('plan/created.xml'):
            plan.save()

        try:
            self.assertEqual(plan.plan_code, plan_code)

            with self.mock_request('plan/exists.xml'):
                same_plan = Plan.get(plan_code)
            self.assertEqual(same_plan.plan_code, plan_code)
            self.assertEqual(same_plan.name, 'Mock Plan')
            self.assertEqual(same_plan.total_billing_cycles, 10)

            plan.plan_interval_length = 2
            plan.plan_interval_unit = 'months'
            plan.unit_amount_in_cents = Money(USD=2000)
            plan.setup_fee_in_cents = Money(USD=200)
            plan.setup_fee_accounting_code = 'Setup Fee AC'
            with self.mock_request('plan/updated.xml'):
                plan.save()
        finally:
            with self.mock_request('plan/deleted.xml'):
                plan.delete()

        """Test taxed plan"""
        with self.mock_request('plan/show-taxed.xml'):
            plan = Plan.get(plan_code)
            self.assertTrue(plan.tax_exempt)

    def test_preview_subscription_change(self):
        with self.mock_request('subscription/show.xml'):
            sub = Subscription.get('123456789012345678901234567890ab')

            with self.mock_request('subscription/change-preview.xml'):
                sub.quantity = 2
                sub.preview()
                self.assertTrue(sub.invoice_collection.charge_invoice.line_items[0].amount_in_cents, 2000)

    def test_subscribe_multiple_errors(self):
        logging.basicConfig(level=logging.DEBUG)  # make sure it's init'ed
        logger = logging.getLogger('recurly.http.request')
        logger.setLevel(logging.DEBUG)

        plan = Plan(
            plan_code='basicplan',
            name='Basic Plan',
            setup_fee_in_cents=Money(0),
            unit_amount_in_cents=Money(1000),
        )
        with self.mock_request('subscription/plan-created.xml'):
            plan.save()

        try:
            account = Account()
            with self.mock_request('subscription/error-subscribe-embedded-account.xml'):
                account.save()
            self.fail("Failed subscription did not raise a Validation error")
        except recurly.ValidationError as err:
            # Check to see that we have a list of suberrors when there is more
            # than one error on a field
            sub_errs = err.errors['subscription.account.account_code']
            self.assertEqual(len(sub_errs), 2)
            self.assertEqual(type(sub_errs[1]), recurly.errors.ValidationError.Suberror)
            self.assertEqual(str(err), "blank: subscription.account.account_code can't be blank; invalid: subscription.account.account_code is invalid; empty: subscription.account.billing_info.address1 can't be empty; empty: subscription.account.billing_info.city can't be empty; empty: subscription.account.billing_info.country can't be empty; blank: subscription.account.billing_info.first_name can't be blank; blank: subscription.account.billing_info.last_name can't be blank; required: subscription.account.billing_info.number is required; empty: subscription.account.billing_info.zip can't be empty; invalid: subscription.plan_code is invalid; not_a_number: subscription.unit_amount_in_cents is not a number")
        except e:
            self.fail("Failed subscription did not raise a Validation error")

    def test_subscribe(self):
        logging.basicConfig(level=logging.DEBUG)  # make sure it's init'ed
        logger = logging.getLogger('recurly.http.request')
        logger.setLevel(logging.DEBUG)

        plan = Plan(
            plan_code='basicplan',
            name='Basic Plan',
            setup_fee_in_cents=Money(0),
            unit_amount_in_cents=Money(1000),
        )
        with self.mock_request('subscription/plan-created.xml'):
            plan.save()

        try:
            account = Account(account_code='subscribe%s' % self.test_id)
            with self.mock_request('subscription/account-created.xml'):
                account.save()

            try:

                sub = Subscription(
                    plan_code='basicplan',
                    currency='USD',
                    unit_amount_in_cents=1000,
                    bulk=True,
                    terms_and_conditions='Some Terms and Conditions',
                    customer_notes='Some Customer Notes',
                    imported_trial=True
                )

                with self.mock_request('subscription/error-no-billing-info.xml'):
                    try:
                        account.subscribe(sub)
                    except BadRequestError as exc:
                        error = exc
                    else:
                        self.fail("Subscribing with no billing info did not raise a BadRequestError")
                self.assertEqual(error.symbol, 'billing_info_required')

                binfo = BillingInfo(
                    first_name='Verena',
                    last_name='Example',
                    address1='123 Main St',
                    city=six.u('San Jos\xe9'),
                    state='CA',
                    zip='94105',
                    country='US',
                    type='credit_card',
                    number='4111 1111 1111 1111',
                    verification_value='7777',
                    year='2015',
                    month='12',
                )
                with self.mock_request('subscription/update-billing-info.xml'):
                    account.update_billing_info(binfo)

                with self.mock_request('subscription/subscribed.xml'):
                    account.subscribe(sub)

                self.assertTrue(sub._url)
                self.assertEquals(sub.imported_trial, True)

                manualsub = Subscription(
                    plan_code='basicplan',
                    currency='USD',
                    net_terms=10,
                    po_number='1000',
                    collection_method='manual'
                )
                with self.mock_request('subscription/subscribed-manual.xml'):
                    account.subscribe(manualsub)
                self.assertTrue(manualsub._url)
                self.assertEqual(manualsub.net_terms, 10)
                self.assertEqual(manualsub.collection_method, 'manual')
                self.assertEqual(manualsub.po_number, '1000')

                shipping_address = ShippingAddress()
                shipping_address.address1 = '123 Main St'
                shipping_address.city = 'San Francisco'
                shipping_address.zip = '94105'
                shipping_address.state = 'CA'
                shipping_address.country = 'US'
                shipping_address.phone = '8015559876'
                shipping_address.nickname = 'Work'

                sub_with_shipping = Subscription(
                    plan_code='basicplan',
                    currency='USD',
                    shipping_address=shipping_address
                )
                with self.mock_request('subscription/subscribed-shipping-address.xml'):
                    account.subscribe(sub_with_shipping)

                with self.mock_request('subscription/account-subscriptions.xml'):
                    subs = account.subscriptions()
                self.assertTrue(len(subs) > 0)
                self.assertEqual(subs[0].uuid, sub.uuid)

                with self.mock_request('subscription/all-subscriptions.xml'):
                    subs = Subscription.all()
                self.assertTrue(len(subs) > 0)
                self.assertEqual(subs[0].uuid, sub.uuid)

                with self.mock_request('subscription/cancelled.xml'):
                    sub.cancel()
                with self.mock_request('subscription/reactivated.xml'):
                    sub.reactivate()

                # Try modifying the subscription.
                sub.timeframe = 'renewal'
                sub.unit_amount_in_cents = 800
                with self.mock_request('subscription/updated-at-renewal.xml'):
                    sub.save()
                pending_sub = sub.pending_subscription
                self.assertTrue(isinstance(pending_sub, Subscription))
                self.assertEqual(pending_sub.unit_amount_in_cents, 800)
                self.assertEqual(sub.unit_amount_in_cents, 1000)

                with self.mock_request('subscription/terminated.xml'):
                    sub.terminate(refund='none')

                log_content = StringIO()
                log_handler = logging.StreamHandler(log_content)
                logger.addHandler(log_handler)

                sub = Subscription(
                    plan_code='basicplan',
                    currency='USD',
                    account=Account(
                        account_code='subscribe%s' % self.test_id,
                        billing_info=BillingInfo(
                            first_name='Verena',
                            last_name='Example',
                            address1='123 Main St',
                            city=six.u('San Jos\xe9'),
                            state='CA',
                            zip='94105',
                            country='US',
                            type='credit_card',
                            number='4111 1111 1111 1111',
                            verification_value='7777',
                            year='2015',
                            month='12',
                        ),
                    ),
                )
                with self.mock_request('subscription/subscribed-billing-info.xml'):
                    account.subscribe(sub)

                logger.removeHandler(log_handler)
                log_content = log_content.getvalue()
                self.assertTrue('<subscription' in log_content)
                self.assertTrue('<billing_info' in log_content)
                # See if we redacted our sensitive fields properly.
                self.assertTrue('4111' not in log_content)
                self.assertTrue('7777' not in log_content)

            finally:
                with self.mock_request('subscription/account-deleted.xml'):
                    account.delete()

            account_code_2 = 'subscribe-%s-2' % self.test_id
            sub = Subscription(
                plan_code='basicplan',
                currency='USD',
                account=Account(
                    account_code=account_code_2,
                    billing_info=BillingInfo(
                        first_name='Verena',
                        last_name='Example',
                        address1='123 Main St',
                        city=six.u('San Jos\xe9'),
                        state='CA',
                        zip='94105',
                        country='US',
                        type='credit_card',
                        number='4111 1111 1111 1111',
                        verification_value='7777',
                        year='2015',
                        month='12',
                    ),
                ),
            )
            with self.mock_request('subscription/subscribe-embedded-account.xml'):
                sub.save()

            with self.mock_request('subscription/embedded-account-exists.xml'):
                acc = Account.get(account_code_2)
            self.assertEqual(acc.account_code, account_code_2)

            with self.mock_request('subscription/embedded-account-deleted.xml'):
                acc.delete()

        finally:
            with self.mock_request('subscription/plan-deleted.xml'):
                plan.delete()

        with self.mock_request('subscription/show.xml'):
            sub = Subscription.get('123456789012345678901234567890ab')
            self.assertEqual(sub.tax_in_cents, 0)
            self.assertEqual(sub.tax_type, 'usst')

            with self.mock_request('subscription/redemptions.xml'):
                self.assertEqual(type(sub.redemptions()), recurly.resource.Page)

    def test_measured_unit(self):
        with self.mock_request('measured-units/exists.xml'):
            measured_unit = MeasuredUnit.get(123456)
            self.assertEqual(measured_unit.name, 'marketing_email')
            self.assertEqual(measured_unit.display_name, 'Marketing Email')
            self.assertEqual(measured_unit.description, 'Unit of Marketing Email')
            self.assertEqual(measured_unit.id, 123456)

    def test_subscription_pause_resume(self):
        with self.mock_request('subscription/show.xml'):
            sub = Subscription.get('123456789012345678901234567890ab')

        with self.mock_request('subscription/pause.xml'):
            sub.pause(1)

        self.assertIsInstance(sub.paused_at, datetime)
        self.assertEqual(sub.remaining_pause_cycles, 1)

        with self.mock_request('subscription/resume.xml'):
            sub.resume()

    def test_usage(self):
        usage = Usage()
        usage.amount = 100 # record 100 emails
        usage.merchant_tag = "Recording 100 emails used by customer"
        usage.recording_timestamp = datetime(2016, 12, 12, 12, 0, 0, 0)
        usage.usage_timestamp = datetime(2016, 12, 12, 12, 0, 0, 0)

        with self.mock_request('subscription/show.xml'):
            sub = Subscription.get('123456789012345678901234567890ab')

            # find the add on with the marketing_emails code
            def marketing_emails_add_on(sub):
                for add_on in sub.subscription_add_ons:
                    if add_on.add_on_code == 'marketing_emails':
                        return add_on
                return None

            add_on = marketing_emails_add_on(sub)

            with self.mock_request('usage/created.xml'):
                sub.create_usage(add_on, usage)

            with self.mock_request('usage/index.xml'):
                usages = add_on.usage()

                self.assertEquals(type(usages), recurly.resource.Page)
                self.assertEquals(len(usages), 1)

                for usage in usages:
                    self.assertEquals(type(usage), Usage)

    def test_subscribe_add_on(self):
        plan = Plan(
            plan_code='basicplan',
            name='Basic Plan',
            setup_fee_in_cents=Money(0),
            unit_amount_in_cents=Money(1000),
        )
        with self.mock_request('subscribe-add-on/plan-created.xml'):
            plan.save()

        try:

            add_on = AddOn(
                add_on_code='mock_add_on',
                name='Mock Add-On',
                unit_amount_in_cents=Money(100),
            )
            with self.mock_request('subscribe-add-on/add-on-created.xml'):
                plan.create_add_on(add_on)

            second_add_on = AddOn(
                add_on_code='second_add_on',
                name='Second Add-On',
                unit_amount_in_cents=Money(50),
            )
            with self.mock_request('subscribe-add-on/second-add-on-created.xml'):
                plan.create_add_on(second_add_on)

            account_code='sad-on-%s' % self.test_id
            sub = Subscription(
                plan_code='basicplan',
                subscription_add_ons=[
                    SubscriptionAddOn(
                        add_on_code='mock_add_on',
                    ),
                    SubscriptionAddOn(
                        add_on_code='second_add_on',
                    ),
                ],
                currency='USD',
                account=Account(
                    account_code=account_code,
                    billing_info=BillingInfo(
                        first_name='Verena',
                        last_name='Example',
                        number='4111 1111 1111 1111',
                        address1='123 Main St',
                        city='San Francisco',
                        state='CA',
                        zip='94105',
                        country='US',
                        verification_value='7777',
                        year='2015',
                        month='12',
                    ),
                ),
            )
            with self.mock_request('subscribe-add-on/subscribed.xml'):
                sub.save()

            # Subscription amounts are in one real currency, so they aren't Money instances.
            sub_amount = sub.unit_amount_in_cents
            self.assertTrue(not isinstance(sub_amount, Money))
            self.assertEqual(sub_amount, 1000)

            # Test that the add-ons' amounts aren't real Money instances either.
            add_on_1, add_on_2 = sub.subscription_add_ons
            self.assertIsInstance(add_on_1, SubscriptionAddOn)
            amount_1 = add_on_1.unit_amount_in_cents
            self.assertTrue(not isinstance(amount_1, Money))
            self.assertEqual(amount_1, 100)

            with self.mock_request('subscribe-add-on/account-exists.xml'):
                account = Account.get(account_code)
            with self.mock_request('subscribe-add-on/account-deleted.xml'):
                account.delete()

        finally:
            with self.mock_request('subscribe-add-on/plan-deleted.xml'):
                plan.delete()

    def test_subscription_notes(self):
        with self.mock_request('subscription/show.xml'):
            sub = Subscription.get('123456789012345678901234567890ab')

        with self.mock_request('subscription/subscribe-notes.xml'):
            sub.terms_and_conditions = "Some terms and conditions"
            sub.customer_notes = "Some customer notes"
            sub.vat_reverse_charge_notes = "Some vat reverse charge notes"
            sub.gateway_code = 'A new gateway code'
            sub.update_notes()

        self.assertEquals(sub.gateway_code, 'A new gateway code')

    def test_subscription_custom_fields(self):
        account_code = 'subscribe-%s-2' % self.test_id
        sub = Subscription(
            plan_code='basicplan',
            currency='USD',
            account=Account(
                account_code=account_code,
                billing_info=BillingInfo(
                    first_name='Verena',
                    last_name='Example',
                    address1='123 Main St',
                    city=six.u('San Jos\xe9'),
                    state='CA',
                    zip='94105',
                    country='US',
                    type='credit_card',
                    number='4111 1111 1111 1111',
                    verification_value='7777',
                    year='2015',
                    month='12',
                ),
                custom_fields=[CustomField(name='my_account_field', value='here is the account value you seek')],
            ),
            custom_fields=[CustomField(name='my_sub_field', value='definitely sub value')],
        )

        with self.mock_request('subscription/subscribe-custom-fields.xml'):
            sub.save()

        self.assertTrue(sub._url)
        self.assertEquals(sub.custom_fields[0].name, 'my_sub_field')
        self.assertEquals(sub.custom_fields[0].value, 'definitely sub value')

        cfs = sub.custom_fields
        cfs[0].value = 'A new sub value'
        sub.custom_fields = cfs

        with self.mock_request('subscription/subscribe-custom-fields-notes.xml'):
            sub.update_notes()

        self.assertEquals(sub.custom_fields[0].value, 'A new sub value')

    def test_account_notes(self):
        account1 = Account(account_code='note%s' % self.test_id)
        account2 = Account(account_code='note%s' % self.test_id)

        with self.mock_request('account-notes/account1-created.xml'):
            account1.save()
        with self.mock_request('account-notes/account2-created.xml'):
            account2.save()
        try:
            with self.mock_request('account-notes/account1-note-list.xml'):
                notes1 = account1.notes()
            with self.mock_request('account-notes/account2-note-list.xml'):
                notes2 = account2.notes()

            # assert accounts don't share notes
            self.assertNotEqual(notes1, notes2)

            # assert contains the proper notes
            self.assertEqual(notes1[0].message, "Python Madness")
            self.assertEqual(notes1[1].message, "Some message")
            self.assertEqual(notes2[0].message, "Foo Bar")
            self.assertEqual(notes2[1].message, "Baz Boo Bop")

        finally:
            with self.mock_request('account-notes/account1-deleted.xml'):
                account1.delete()
            with self.mock_request('account-notes/account2-deleted.xml'):
                account2.delete()

    def test_transaction(self):
        logging.basicConfig(level=logging.DEBUG)  # make sure it's init'ed
        logger = logging.getLogger('recurly.http.request')
        logger.setLevel(logging.DEBUG)

        account_code = 'transaction%s' % self.test_id

        log_content = StringIO()
        log_handler = logging.StreamHandler(log_content)
        logger.addHandler(log_handler)

        transaction = Transaction(
            amount_in_cents=1000,
            currency='USD',
            account=Account(
                account_code=account_code,
                billing_info=BillingInfo(
                    first_name='Verena',
                    last_name='Example',
                    number='4111-1111-1111-1111',
                    year='2014',
                    address1='123 Main St',
                    city='San Francisco',
                    state='CA',
                    zip='94105',
                    country='US',
                    month='7',
                    verification_value='7777',
                ),
            )
        )
        with self.mock_request('transaction/created.xml'):
            transaction.save()

        fraud_info = transaction.fraud
        self.assertEquals(fraud_info.score, 88)
        self.assertEquals(fraud_info.decision, 'DECLINED')

        logger.removeHandler(log_handler)

        try:
            transaction.get_refund_transaction()
        except ValueError:
            pass
        else:
            self.fail("Transaction with no refund transaction did not raise a ValueError from get_refund_transaction()")

        with self.mock_request('transaction/account-exists.xml'):
            account = Account.get(account_code)

        try:
            log_content = log_content.getvalue()
            self.assertTrue('<transaction' in log_content)
            self.assertTrue('<billing_info' in log_content)
            # See if we redacted our sensitive fields properly.
            self.assertTrue('4111' not in log_content)
            self.assertTrue('7777' not in log_content)

            with self.mock_request('transaction/refunded.xml'):
                refunded_transaction = transaction.refund()

            transaction_2 = Transaction(
                amount_in_cents=1000,
                currency='USD',
                account=Account(account_code=account_code),
            )
            with self.mock_request('transaction/created-again.xml'):
                transaction_2.save()
            self.assertNotEqual(transaction_2.uuid, transaction.uuid)
            self.assertTrue(transaction_2.refundable)

            with self.mock_request('transaction/partial-refunded.xml'):
                refunded_transaction = transaction_2.refund(amount_in_cents=700)
            self.assertTrue(refunded_transaction is transaction_2)
            self.assertTrue(hasattr(transaction_2, 'get_refund_transaction'))
            with self.mock_request('transaction/partial-refunded-transaction.xml'):
                refund_transaction = transaction_2.get_refund_transaction()
            self.assertTrue(isinstance(refund_transaction, Transaction))
            self.assertTrue(not refund_transaction.refundable)
            self.assertNotEqual(refund_transaction.uuid, transaction_2.uuid)

        finally:
            with self.mock_request('transaction/account-deleted.xml'):
                account.delete()

    def failed_transaction(self):
        transaction = Transaction(
            amount_in_cents=1000,
            currency='USD',
            account=Account(),
        )

        with self.mock_request('transaction/declined-stransaction.xml'):
            try:
                transaction.save()
            except ValidationError as _error:
                error = _error
            else:
                self.fail("Posting a transaction without an account code did not raise a ValidationError")

        self.assertEqual(transaction.error_code, 'insufficient_funds')
        self.assertEqual(transaction.error_category, 'soft')
        self.assertEqual(transaction.customer_message, 'The transaction was declined due to insufficient funds in your account. Please use a different card or contact your bank.')
        self.assertEqual(transaction.merchant_message, 'The card has insufficient funds to cover the cost of the transaction.')
        self.assertEqual(transaction.gateway_error_code, '123')


    def test_transaction_with_balance(self):
        transaction = Transaction(
            amount_in_cents=1000,
            currency='USD',
            account=Account(),
        )
        error = None
        with self.mock_request('transaction-balance/transaction-no-account.xml'):
            try:
                transaction.save()
            except ValidationError as _error:
                error = _error
            else:
                self.fail("Posting a transaction without an account code did not raise a ValidationError")
        # Make sure there really were errors.
        self.assertTrue(len(error.errors) > 0)

        account_code = 'transbalance%s' % self.test_id
        account = Account(account_code=account_code)
        with self.mock_request('transaction-balance/account-created.xml'):
            account.save()

        try:
            # Try to charge without billing info, should break.
            transaction = Transaction(
                amount_in_cents=1000,
                currency='USD',
                account=account,
            )
            error = None
            with self.mock_request('transaction-balance/transaction-no-billing-fails.xml'):
                try:
                    transaction.save()
                except ValidationError as _error:
                    error = _error
                else:
                    self.fail("Posting a transaction without billing info did not raise a ValidationError")
            # Make sure there really were errors.
            self.assertTrue(len(error.errors) > 0)

            binfo = BillingInfo(
                first_name='Verena',
                last_name='Example',
                address1='123 Main St',
                city=six.u('San Jos\xe9'),
                state='CA',
                zip='94105',
                country='US',
                type='credit_card',
                number='4111 1111 1111 1111',
                verification_value='7777',
                year='2015',
                month='12',
            )
            with self.mock_request('transaction-balance/set-billing-info.xml'):
                account.update_billing_info(binfo)

            # Try to charge now, should be okay.
            transaction = Transaction(
                amount_in_cents=1000,
                currency='USD',
                account=account,
            )
            with self.mock_request('transaction-balance/transacted.xml'):
                transaction.save()

            # Give the account a credit.
            credit = Adjustment(unit_amount_in_cents=-2000, currency='USD', description='transaction test credit')
            with self.mock_request('transaction-balance/credited.xml'):
                # TODO: maybe this should be adjust()?
                account.charge(credit)

            # Try to charge less than the account balance, which should fail (not a CC transaction).
            transaction = Transaction(
                amount_in_cents=500,
                currency='USD',
                account=account,
            )
            with self.mock_request('transaction-balance/transacted-2.xml'):
                transaction.save()
            # The transaction doesn't actually save.
            self.assertTrue(transaction._url is None)

            # Try to charge more than the account balance, which should work.
            transaction = Transaction(
                amount_in_cents=3000,
                currency='USD',
                account=account,
            )
            with self.mock_request('transaction-balance/transacted-3.xml'):
                transaction.save()
            # This transaction should be recorded.
            self.assertTrue(transaction._url is not None)

        finally:
            with self.mock_request('transaction-balance/account-deleted.xml'):
                account.delete()

    def _build_gift_card(self):
        account_code = 'e0004e3c-216c-4254-8767-9be605cd0b03'
        account = recurly.Account(account_code=account_code)
        account.email = 'verena@example.com'
        account.first_name = 'Verena'
        account.last_name = 'Example'

        billing_info = BillingInfo()
        billing_info.first_name = 'Verena'
        billing_info.last_name = 'Example'
        billing_info.number = '4111-1111-1111-1111'
        billing_info.verification_value = '123'
        billing_info.month = 11
        billing_info.year = 2019
        billing_info.country = 'US'

        address = Address()
        address.address1 = '400 Alabama St'
        address.zip = '94110'
        address.city = 'San Francisco'
        address.state = 'CA'
        address.country = 'US'

        delivery = Delivery()
        delivery.method = 'email'
        delivery.email_address = 'john@email.com'
        delivery.first_name = 'John'
        delivery.last_name = 'Smith'

        gift_card = GiftCard()
        gift_card.product_code = 'test_gift_card'
        gift_card.currency = 'USD'
        gift_card.unit_amount_in_cents = 2000

        delivery.address = address
        account.billing_info = billing_info

        gift_card.delivery = delivery
        gift_card.gifter_account = account
        return gift_card

    def test_gift_cards_purchase(self):
        gift_card = self._build_gift_card()

        # now allowed to send a top-level billing info along
        gift_card.billing_info = BillingInfo(token_id='1234')

        self.assertFalse('_url' in gift_card.attributes)

        with self.mock_request('gift_cards/created.xml'):
            gift_card.save()

        self.assertTrue(gift_card._url is not None)
        self.assertTrue(gift_card.delivery is not None)
        self.assertTrue(gift_card.canceled_at is None)

    def test_gift_cards_preview(self):
        gift_card = self._build_gift_card()

        self.assertFalse('_url' in gift_card.attributes)

        with self.mock_request('gift_cards/preview.xml'):
            gift_card.preview()

        self.assertTrue(gift_card.id is None)
        self.assertFalse('_url' in gift_card.attributes)

    def test_gift_cards_redeem(self):
        gift_card = GiftCard(redemption_code='9FC359369CD3892E')

        with self.mock_request('gift_cards/redeem.xml'):
            gift_card.redeem('e0004e3c-216c-4254-8767-9be605cd0b03')

        self.assertTrue(gift_card.redeemed_at is not None)

    def test_gift_cards_redeem_with_url(self):
        gift_card = GiftCard(redemption_code='9FC359369CD3892E')
        gift_card._url = 'https://api.recurly.com/v2/gift_cards/9FC359369CD3892E'

        with self.mock_request('gift_cards/redeem.xml'):
            gift_card.redeem('e0004e3c-216c-4254-8767-9be605cd0b03')

        self.assertTrue(gift_card.redeemed_at is not None)

    def test_export_date(self):
        with self.mock_request('export-date/export-date.xml'):
            export_dates = ExportDate.all()

        self.assertEqual(len(export_dates), 1)
        self.assertEqual(export_dates[0].date, "2019-05-09")

    def test_export_date_files(self):
        export_date = ExportDate()

        with self.mock_request('export-date-files/export-date-files-list.xml'):
            export_date_files = export_date.files("2019-05-09")

        self.assertEqual(len(export_date_files), 1)
        self.assertEqual(export_date_files[0].name, "churned_subscriptions_v2_expires.csv.gz")

    def test_export_date_files_download_information(self):
        export_date = ExportDate()

        with self.mock_request('export-date-files/export-date-files-list.xml'):
            export_date_files = export_date.files("2019-05-09")

        with self.mock_request('export-date-files/export-date-file-download-information.xml'):
            export_date_file_download_information = export_date_files[0].download_information()

        self.assertEqual(
            export_date_file_download_information.expires_at.strftime("%Y-%m-%d %H:%M:%S"), "2019-05-09 14:00:00"
        )
        self.assertEqual(export_date_file_download_information.download_url, "https://api.recurly.com/download")


if __name__ == '__main__':
    import unittest
    unittest.main()



from recurly import recurly_logging as logging
import sys
import re
from datetime import datetime
import six
from six.moves.urllib.parse import urljoin
from six import iteritems
from defusedxml import ElementTree
from xml.etree import ElementTree as ElementTreeBuilder

import recurly
from recurly.errors import *
from recurly.resource import Resource, Money, PageError, Page

"""

Recurly's Python client library is an interface to its REST API.

Please see the Recurly API documentation for more information:

https://dev.recurly.com/docs/getting-started

"""

__version__ = '2.9.10'
__python_version__ = '.'.join(map(str, sys.version_info[:3]))

cached_rate_limits = {
        'limit': None,
        'remaining': None,
        'resets_at': None,
        'cached_at': None,
        }

VALID_DOMAINS = ('.recurly.com',)
"""A tuple of whitelisted domains that this client can connect to."""

USER_AGENT = 'recurly-python/%s; python %s; %s' % (recurly.__version__, recurly.__python_version__, recurly.resource.ssl.OPENSSL_VERSION)

BASE_URI = 'https://%s.recurly.com/v2/'
"""The API endpoint to send requests to."""

SUBDOMAIN = 'api'
"""The subdomain of the site authenticating API requests."""

API_KEY = None
"""The API key to use when authenticating API requests."""

API_VERSION = '2.22'
"""The API version to use when making API requests."""

CA_CERTS_FILE = None
"""A file contianing a set of concatenated certificate authority certs
for validating the server against."""

DEFAULT_CURRENCY = 'USD'
"""The currency to use creating `Money` instances when one is not specified."""

SOCKET_TIMEOUT_SECONDS = None
"""The number of seconds after which to timeout requests to the Recurly API.
If unspecified, the global default timeout is used."""

def urljoin(url1, url2):
    if url1.endswith('/'):
        url1 = url1[0:-1]
    if not url2.startswith('/'):
        url2 = '/' + url2
    return url1 + url2

def base_uri():
    if SUBDOMAIN is None:
        raise ValueError('recurly.SUBDOMAIN not set')

    return BASE_URI % SUBDOMAIN

def api_version():
    return API_VERSION

def cache_rate_limit_headers(resp_headers):
    try:
        recurly.cached_rate_limits = {
                'cached_at': datetime.utcnow(),
                'limit': int(resp_headers['X-RateLimit-Limit']),
                'remaining': int(resp_headers['X-RateLimit-Remaining']),
                'resets_at': datetime.utcfromtimestamp(int(resp_headers['X-RateLimit-Reset']))
                }
    except:
        log = logging.getLogger('recurly.cached_rate_limits')
        log.info('Failed to parse rate limits from header')

class Address(Resource):

    nodename = 'address'

    attributes = (
        'first_name',
        'last_name',
        'name_on_account',
        'company',
        'address1',
        'address2',
        'city',
        'state',
        'zip',
        'country',
        'phone',
    )

class AccountBalance(Resource):

    """The Balance on an account"""

    nodename = 'account_balance'

    attributes = (
        'balance_in_cents',
        'past_due',
    )

class AccountAcquisition(Resource):

    """Account acquisition data https://dev.recurly.com/docs/create-account-acquisition"""

    nodename = 'account_acquisition'

    attributes = (
        'cost_in_cents',
        'currency',
        'channel',
        'subchannel',
        'campaign',
        'created_at',
        'updated_at',
    )

class CustomField(Resource):

    """A field to store extra data on the account or subscription."""

    nodename = 'custom_field'

    attributes = (
        'name',
        'value',
    )

    def to_element(self, root_name=None):
        # Include the field name/value pair when the value changed
        if 'value' in self.__dict__:
            try:
                self.name = self.name # forces name into __dict__
            except AttributeError:
                pass

        return super(CustomField, self).to_element(root_name)

class Account(Resource):

    """A customer account."""

    member_path = 'accounts/%s'
    collection_path = 'accounts'

    nodename = 'account'

    attributes = (
        'account_code',
        'parent_account_code',
        'username',
        'email',
        'first_name',
        'last_name',
        'company_name',
        'vat_number',
        'tax_exempt',
        'exemption_certificate',
        'entity_use_code',
        'accept_language',
        'cc_emails',
        'account_balance',
        'created_at',
        'updated_at',
        'shipping_addresses',
        'account_acquisition',
        'has_live_subscription',
        'has_active_subscription',
        'has_future_subscription',
        'has_canceled_subscription',
        'has_paused_subscription',
        'has_past_due_invoice',
        'preferred_locale',
        'custom_fields',
        'transaction_type',
    )

    _classes_for_nodename = { 'address': Address, 'custom_field': CustomField }

    sensitive_attributes = ('number', 'verification_value',)

    def to_element(self, root_name=None):
        elem = super(Account, self).to_element(root_name)

        # Make sure the account code is always included in a serialization.
        if 'account_code' not in self.__dict__:  # not already included
            try:
                account_code = self.account_code
            except AttributeError:
                pass
            else:
                elem.append(self.element_for_value('account_code', account_code))
        if 'billing_info' in self.__dict__:
            elem.append(self.billing_info.to_element())
        if 'address' in self.__dict__:
            elem.append(self.address.to_element())

        return elem

    @classmethod
    def all_active(cls, **kwargs):
        """Return a `Page` of active customer accounts.

        This is a convenience method for `Account.all(state='active')`.

        """
        return cls.all(state='active', **kwargs)

    @classmethod
    def all_closed(cls, **kwargs):
        """Return a `Page` of closed customer accounts.

        This is a convenience method for `Account.all(state='closed')`.

        """
        return cls.all(state='closed', **kwargs)

    @classmethod
    def all_past_due(cls, **kwargs):
        """Return a `Page` of past-due customer accounts.

        This is a convenience method for `Account.all(state='past_due').

        """
        return cls.all(state='past_due', **kwargs)

    @classmethod
    def all_subscribers(cls, **kwargs):
        """Return a `Page` of customer accounts that are subscribers.

        This is a convenience method for `Account.all(state='subscriber').

        """
        return cls.all(state='subscriber', **kwargs)

    @classmethod
    def all_non_subscribers(cls, **kwargs):
        """Return a `Page` of customer accounts that are not subscribers.

        This is a convenience method for `Account.all(state='non_subscriber').

        """
        return cls.all(state='non_subscriber', **kwargs)

    def __getattr__(self, name):
        if name == 'billing_info':
            try:
                billing_info_url = self._elem.find('billing_info').attrib['href']
            except (AttributeError, KeyError):
                raise AttributeError(name)
            resp, elem = BillingInfo.element_for_url(billing_info_url)
            return BillingInfo.from_element(elem)
        try:
            return super(Account, self).__getattr__(name)
        except AttributeError:
            if name == 'address':
                self.address = Address()
                return self.address
            else:
              raise AttributeError(name)

    def charge(self, charge):
        """Charge (or credit) this account with the given `Adjustment`."""
        url = urljoin(self._url, '/adjustments')
        return charge.post(url)

    def invoice(self, **kwargs):
        """Create an invoice for any outstanding adjustments this account has."""
        url = urljoin(self._url, '/invoices')

        if kwargs:
            response = self.http_request(url, 'POST', Invoice(**kwargs), {'Content-Type':
                'application/xml; charset=utf-8'})
        else:
            response = self.http_request(url, 'POST')

        if response.status != 201:
            self.raise_http_error(response)

        response_xml = response.read()
        logging.getLogger('recurly.http.response').debug(response_xml)
        elem = ElementTree.fromstring(response_xml)

        invoice_collection = InvoiceCollection.from_element(elem)
        return invoice_collection

    def build_invoice(self):
        """Preview an invoice for any outstanding adjustments this account has."""
        url = urljoin(self._url, '/invoices/preview')

        response = self.http_request(url, 'POST')
        if response.status != 200:
            self.raise_http_error(response)

        response_xml = response.read()
        logging.getLogger('recurly.http.response').debug(response_xml)
        elem = ElementTree.fromstring(response_xml)

        invoice_collection = InvoiceCollection.from_element(elem)
        return invoice_collection

    def notes(self):
        """Fetch Notes for this account."""
        url = urljoin(self._url, '/notes')
        return Note.paginated(url)

    def redemption(self):
      try:
        return self.redemptions()[0]
      except AttributeError:
        raise AttributeError("redemption")

    def reopen(self):
        """Reopen a closed account."""
        url = urljoin(self._url, '/reopen')
        response = self.http_request(url, 'PUT')
        if response.status != 200:
            self.raise_http_error(response)

        response_xml = response.read()
        logging.getLogger('recurly.http.response').debug(response_xml)
        self.update_from_element(ElementTree.fromstring(response_xml))

    def subscribe(self, subscription):
        """Create the given `Subscription` for this existing account."""
        url = urljoin(self._url, '/subscriptions')
        return subscription.post(url)

    def update_billing_info(self, billing_info):
        """Change this account's billing information to the given `BillingInfo`."""
        url = urljoin(self._url, '/billing_info')
        response = billing_info.http_request(url, 'PUT', billing_info,
            {'Content-Type': 'application/xml; charset=utf-8'})
        if response.status == 200:
            pass
        elif response.status == 201:
            billing_info._url = response.getheader('Location')
        else:
            billing_info.raise_http_error(response)

        response_xml = response.read()
        logging.getLogger('recurly.http.response').debug(response_xml)
        billing_info.update_from_element(ElementTree.fromstring(response_xml))

    def create_shipping_address(self, shipping_address):
        """Creates a shipping address on an existing account. If you are
        creating an account, you can embed the shipping addresses with the
        request"""
        url = urljoin(self._url, '/shipping_addresses')
        return shipping_address.post(url)

class BillingInfo(Resource):

    """A set of billing information for an account."""

    nodename = 'billing_info'

    attributes = (
        'type',
        'name_on_account',
        'first_name',
        'last_name',
        'number',
        'verification_value',
        'year',
        'month',
        'start_month',
        'start_year',
        'issue_number',
        'company',
        'address1',
        'address2',
        'city',
        'state',
        'zip',
        'country',
        'phone',
        'vat_number',
        'ip_address',
        'ip_address_country',
        'card_type',
        'first_six',
        'last_four',
        'paypal_billing_agreement_id',
        'amazon_billing_agreement_id',
        'amazon_region',
        'token_id',
        'account_type',
        'routing_number',
        'account_number',
        'currency',
        'updated_at',
        'external_hpp_type',
        'gateway_token',
        'gateway_code',
        'three_d_secure_action_result_token_id',
        'transaction_type',
    )
    sensitive_attributes = ('number', 'verification_value', 'account_number')
    xml_attribute_attributes = ('type',)

class ShippingAddress(Resource):

    """Shipping Address information"""

    nodename = 'shipping_address'

    attributes = (
        'address1',
        'address2',
        'city',
        'company',
        'country',
        'email',
        'first_name',
        'id',
        'last_name',
        'nickname',
        'phone',
        'state',
        'zip',
    )

class Delivery(Resource):

    """Delivery information for use with a Gift Card"""

    nodename = 'delivery'

    attributes = (
        'address',
        'deliver_at',
        'email_address',
        'first_name',
        'gifter_name',
        'last_name',
        'method',
        'personal_message',
    )

# This is used internally for proper XML generation
class _RecipientAccount(Account):
    nodename = 'recipient_account'

class GiftCard(Resource):

    """A Gift Card for a customer to purchase or apply to a subscription or account."""

    member_path= 'gift_cards/%s'
    collection_path = 'gift_cards'

    nodename = 'gift_card'

    attributes = (
        'balance_in_cents',
        'canceled_at',
        'created_at',
        'currency',
        'delivery',
        'gifter_account',
        'id',
        'invoice',
        'product_code',
        'recipient_account',
        'redeemed_at',
        'redemption_code',
        'updated_at',
        'unit_amount_in_cents',
        'billing_info',
    )
    _classes_for_nodename = {'recipient_account': Account,'gifter_account':
            Account, 'delivery': Delivery}

    def preview(self):
        """Preview the purchase of this gift card"""

        if hasattr(self, '_url'):
            url = self._url + '/preview'
            return self.post(url)
        else:
            url = urljoin(recurly.base_uri(), self.collection_path + '/preview')
            return self.post(url)

    def redeem(self, account_code):
        """Redeem this gift card on the specified account code"""

        redemption_path = '%s/redeem' % (self.redemption_code)

        if hasattr(self, '_url'):
            url = urljoin(self._url, '/redeem')
        else:
            url = urljoin(recurly.base_uri(), self.collection_path + '/' + redemption_path)

        recipient_account = _RecipientAccount(account_code=account_code)
        return self.post(url, recipient_account)

    def to_element(self, root_name=None):
        elem = super(GiftCard, self).to_element(root_name)

        # Make sure the redemption code is always included in a serialization.
        if 'redemption_code' not in self.__dict__:  # not already included
            try:
                redemption_code = self.redemption_code
            except AttributeError:
                pass
            else:
                elem.append(self.element_for_value('redemption_code',
                    redemption_code))

        return elem

class Coupon(Resource):

    """A coupon for a customer to apply to their account."""

    member_path = 'coupons/%s'
    collection_path = 'coupons'

    nodename = 'coupon'

    attributes = (
        'coupon_code',
        'name',
        'discount_type',
        'discount_percent',
        'discount_in_cents',
        'redeem_by_date',
        'invoice_description',
        'single_use',
        'applies_for_months',
        'duration',
        'temporal_unit',
        'temporal_amount',
        'max_redemptions',
        'applies_to_all_plans',
        'applies_to_non_plan_charges',
        'redemption_resource',
        'created_at',
        'updated_at',
        'deleted_at',
        'plan_codes',
        'hosted_description',
        'max_redemptions_per_account',
        'coupon_type',
        'unique_code_template',
        'unique_coupon_codes',
        'free_trial_unit',
        'free_trial_amount',
        'id',
    )

    @classmethod
    def value_for_element(cls, elem):
        if elem is None or elem.tag != 'plan_codes' or elem.attrib.get('type') != 'array':
            return super(Coupon, cls).value_for_element(elem)

        return [code_elem.text for code_elem in elem]

    @classmethod
    def element_for_value(cls, attrname, value):
        if attrname != 'plan_codes':
            return super(Coupon, cls).element_for_value(attrname, value)

        elem = ElementTreeBuilder.Element(attrname)
        elem.attrib['type'] = 'array'
        for code in value:
            code_el = ElementTreeBuilder.Element('plan_code')
            code_el.text = code
            elem.append(code_el)

        return elem

    @classmethod
    def all_redeemable(cls, **kwargs):
        """Return a `Page` of redeemable coupons.

        This is a convenience method for `Coupon.all(state='redeemable')`.

        """
        return cls.all(state='redeemable', **kwargs)

    @classmethod
    def all_expired(cls, **kwargs):
        """Return a `Page` of expired coupons.

        This is a convenience method for `Coupon.all(state='expired')`.

        """
        return cls.all(state='expired', **kwargs)

    @classmethod
    def all_maxed_out(cls, **kwargs):
        """Return a `Page` of coupons that have been used the maximum
        number of times.

        This is a convenience method for `Coupon.all(state='maxed_out')`.

        """
        return cls.all(state='maxed_out', **kwargs)

    def has_unlimited_redemptions_per_account(self):
        return self.max_redemptions_per_account == None

    def generate(self, amount):
        elem = ElementTreeBuilder.Element(self.nodename)
        elem.append(Resource.element_for_value('number_of_unique_codes', amount))

        url = urljoin(self._url, '/generate')
        body = ElementTree.tostring(elem, encoding='UTF-8')

        response = self.http_request(url, 'POST', body, { 'Content-Type':
            'application/xml; charset=utf-8' })

        if response.status not in (200, 201, 204):
            self.raise_http_error(response)

        return Page.page_for_url(response.getheader('Location'))

    def restore(self):
        url = urljoin(self._url, '/restore')
        self.put(url)

class Redemption(Resource):

    """A particular application of a coupon to a customer account."""

    nodename = 'redemption'

    attributes = (
        'account_code',
        'single_use',
        'total_discounted_in_cents',
        'subscription_uuid',
        'currency',
        'created_at',
        'updated_at',
    )

class TaxDetail(Resource):

    """A charge's tax breakdown"""

    nodename = 'taxdetail'
    inherits_currency = True

    attributes = (
        'name',
        'type',
        'tax_rate',
        'tax_in_cents',
    )

class Adjustment(Resource):

    """A charge or credit applied (or to be applied) to an account's invoice."""

    nodename = 'adjustment'
    member_path = 'adjustments/%s'

    attributes = (
        'uuid',
        'description',
        'accounting_code',
        'product_code',
        'quantity',
        'unit_amount_in_cents',
        'discount_in_cents',
        'tax_in_cents',
        'tax_type',
        'tax_region',
        'tax_rate',
        'total_in_cents',
        'currency',
        'tax_exempt',
        'tax_code',
        'tax_details',
        'start_date',
        'end_date',
        'created_at',
        'updated_at',
        'type',
        'revenue_schedule_type',
        'shipping_address',
        'shipping_address_id',
    )
    xml_attribute_attributes = ('type',)
    _classes_for_nodename = {'tax_detail': TaxDetail, 'shipping_address':
                             ShippingAddress}

    # This can be removed when the `original_adjustment_uuid` is moved to a link
    def __getattr__(self, name):
        if name == 'original_adjustment':
            try:
                uuid = super(Adjustment, self).__getattr__('original_adjustment_uuid')
            except (AttributeError):
                return super(Adjustment, self).__getattr__(name)

            return lambda: Adjustment.get(uuid)
        else:
            return super(Adjustment, self).__getattr__(name)

class Invoice(Resource):

    """A payable charge to an account for the customer's charges and
    subscriptions."""

    member_path = 'invoices/%s'
    collection_path = 'invoices'

    nodename = 'invoice'

    attributes = (
        'uuid',
        'state',
        'invoice_number',
        'invoice_number_prefix',
        'po_number',
        'vat_number',
        'tax_in_cents',
        'tax_type',
        'tax_rate',
        'total_in_cents',
        'currency',
        'created_at',
        'updated_at',
        'line_items',
        'transactions',
        'terms_and_conditions',
        'customer_notes',
        'vat_reverse_charge_notes', # Only shows if reverse charge invoice
        'address',
        'closed_at',
        'collection_method',
        'net_terms',
        'attempt_next_collection_at',
        'recovery_reason',
        'balance_in_cents',
        'subtotal_before_discount_in_cents',
        'subtotal_in_cents',
        'discount_in_cents',
        'due_on',
        'type',
        'origin',
        'credit_customer_notes',
        'gateway_code',
        'billing_info',
    )

    blacklist_attributes = (
        'currency',
    )

    def invoice_number_with_prefix(self):
        return '%s%s' % (self.invoice_number_prefix, self.invoice_number)

    @classmethod
    def all_pending(cls, **kwargs):
        """Return a `Page` of pending invoices.

        This is a convenience method for `Invoice.all(state='pending')`.

        """
        return cls.all(state='pending', **kwargs)

    @classmethod
    def all_paid(cls, **kwargs):
        """Return a `Page` of paid invoices.

        This is a convenience method for `Invoice.all(state='paid')`.

        """
        return cls.all(state='paid', **kwargs)

    @classmethod
    def all_failed(cls, **kwargs):
        """Return a `Page` of failed invoices.

        This is a convenience method for `Invoice.all(state='failed')`.

        """
        return cls.all(state='failed', **kwargs)

    @classmethod
    def all_past_due(cls, **kwargs):
        """Return a `Page` of past-due invoices.

        This is a convenience method for `Invoice.all(state='past_due')`.

        """
        return cls.all(state='past_due', **kwargs)

    @classmethod
    def pdf(cls, uuid):
        """Return a PDF of the invoice identified by the UUID

        This is a raw string, which can be written to a file with:
        `
            with open('invoice.pdf', 'w') as invoice_file:
                invoice_file.write(recurly.Invoice.pdf(uuid))
        `

        """
        url = urljoin(base_uri(), cls.member_path % (uuid,))
        pdf_response = cls.http_request(url, headers={'Accept': 'application/pdf'})
        return pdf_response.read()

    def refund_amount(self, amount_in_cents, refund_options = {}):
        # For backwards compatibility
        # TODO the consequent branch of this conditional should eventually be removed
        # and we should document that as a breaking change in the changelog.
        # The same change should be applied to the refund() method
        if (isinstance(refund_options, six.string_types)):
            refund_options = { 'refund_method': refund_options }
        else:
            if 'refund_method' not in refund_options:
                refund_options = { 'refund_method': 'credit_first' }

        amount_element = self._refund_open_amount_xml(amount_in_cents,
                                                     refund_options)
        return self._create_refund_invoice(amount_element)

    def refund(self, adjustments, refund_options = {}):
       # For backwards compatibility
       # TODO the consequent branch of this conditional should eventually be removed
       # and we should document that as a breaking change in the changelog.
       # The same change should be applied to the refund_amount() method
       if (isinstance(refund_options, six.string_types)):
           refund_options = { 'refund_method': refund_options }
       else:
           if 'refund_method' not in refund_options:
               refund_options = { 'refund_method': 'credit_first' }

       adjustments_element = self._refund_line_items_xml(adjustments,
                                                          refund_options)
       return self._create_refund_invoice(adjustments_element)

    def _refund_open_amount_xml(self, amount_in_cents, refund_options):
        elem = ElementTreeBuilder.Element(self.nodename)
        elem.append(Resource.element_for_value('amount_in_cents',
            amount_in_cents))

        # Need to sort the keys for tests to pass in python 2 and 3
        # Can remove `sorted` when we drop python 2 support
        for k, v in sorted(iteritems(refund_options)):
            elem.append(Resource.element_for_value(k, v))
        return elem

    def _refund_line_items_xml(self, line_items, refund_options):
        elem = ElementTreeBuilder.Element(self.nodename)

        line_items_elem = ElementTreeBuilder.Element('line_items')

        for item in line_items:
            adj_elem = ElementTreeBuilder.Element('adjustment')
            adj_elem.append(Resource.element_for_value('uuid',
                item['adjustment'].uuid))
            adj_elem.append(Resource.element_for_value('quantity',
            item['quantity']))
            adj_elem.append(Resource.element_for_value('prorate', item['prorate']))
            line_items_elem.append(adj_elem)

        elem.append(line_items_elem)

        # Need to sort the keys for tests to pass in python 2 and 3
        # Can remove `sorted` when we drop python 2 support
        for k, v in sorted(iteritems(refund_options)):
            elem.append(Resource.element_for_value(k, v))

        return elem

    def mark_failed(self):
        url = urljoin(self._url, '/mark_failed')

        collection = InvoiceCollection()
        response = self.http_request(url, 'PUT')
        if response.status != 200:
            self.raise_http_error(response)
        response_xml = response.read()
        logging.getLogger('recurly.http.response').debug(response_xml)
        collection.update_from_element(ElementTree.fromstring(response_xml))

        return collection

    def _create_refund_invoice(self, element):
        url = urljoin(self._url, '/refund')
        body = ElementTree.tostring(element, encoding='UTF-8')

        refund_invoice = Invoice()
        refund_invoice.post(url, body)

        return refund_invoice

    def redemption(self):
      try:
        return self.redemptions()[0]
      except AttributeError:
        raise AttributeError("redemption")

    def enter_offline_payment(self, transaction):
        """
        Records an offline (external) payment on the invoice.
        Pass in a Transaction object to set the details of the created
        transaction. The attributes available to set are
        (payment_method, collected_at, amount_in_cents, description)

        Returns:
            Transaction: The created transaction
        """
        url = urljoin(self._url, '/transactions')
        transaction.post(url)
        return transaction

    def save(self):
        if hasattr(self, '_url'):
            super(Invoice, self).save()
        else:
            raise BadRequestError("New invoices cannot be created using Invoice#save")

class InvoiceCollection(Resource):

    """A collection of invoices resulting from some action. Includes
    a charge invoice and a list of credit invoices.
    """

    nodename = 'invoice_collection'
    attributes = (
        'charge_invoice',
        'credit_invoices',
    )
    _classes_for_nodename = {
        'charge_invoice': Invoice,
        'credit_invoice': Invoice,
    }

class Purchase(Resource):

    """
    A purchase allows the programmer to invoice multiple subscriptions,
    adjustments, coupon codes, and gift cards all in one call.
    """

    collection_path = 'purchases'
    nodename = 'purchase'

    attributes = (
        'account',
        'adjustments',
        'currency',
        'po_number',
        'net_terms',
        'gift_card',
        'coupon_codes',
        'subscriptions',
        'customer_notes',
        'terms_and_conditions',
        'vat_reverse_charge_notes',
        'shipping_address_id',
        'shipping_fees',
        'gateway_code',
        'collection_method',
        'transaction_type',
    )

    def invoice(self):
        """
        Will invoice the purchase object and run all validations and transactions.

        Returns:
            InvoiceCollection: The generated collection of invoices
        """
        return self.__invoice(self.collection_path)

    def preview(self):
        """
        Will create a preview invoice for the purchase. It runs all the
        validations but not the transactions.

        Returns:
            InvoiceCollection: The preview of collection of invoices
        """
        return self.__invoice(self.collection_path + '/preview')

    def authorize(self):
        """
        Will generate an authorized invoice for the purchase. Runs validations
        but does not run any transactions. This endpoint will create a
        pending purchase that can be activated at a later time once payment
        has been completed on an external source (e.g. Adyen's Hosted
        Payment Pages).

        Returns:
            InvoiceCollection: The authorized collection of invoices
        """
        return self.__invoice(self.collection_path + '/authorize')

    def pending(self):
        """
        Use for Adyen HPP transaction requests. Runs validations
        but does not run any transactions.

        Returns:
            InvoiceCollection: The pending collection of invoices
        """
        return self.__invoice(self.collection_path + '/pending')

    def __invoice(self, url):
        # We must null out currency in subscriptions and adjustments
        # TODO we should deprecate and remove default currency support
        def filter_currency(resources):
            for resource in resources:
                resource.attributes = tuple([a for a in resource.attributes if
                                             a != 'currency'])
        try:
            filter_currency(self.adjustments)
        except AttributeError:
            pass
        try:
            filter_currency(self.subscriptions)
        except AttributeError:
            pass

        url = urljoin(recurly.base_uri(), url)
        response = self.http_request(url, 'POST', self)
        if response.status not in (200, 201):
            self.raise_http_error(response)
        response_xml = response.read()
        logging.getLogger('recurly.http.response').debug(response_xml)
        elem = ElementTree.fromstring(response_xml)
        invoice_collection = InvoiceCollection.from_element(elem)
        return invoice_collection

class ShippingFee(Resource):

    """A one time shipping fee on a Purchase"""

    nodename = 'shipping_fee'

    attributes = (
        'shipping_method_code',
        'shipping_amount_in_cents',
        'shipping_address',
        'shipping_address_id',
    )

class ShippingMethod(Resource):

    """A shipping method available on the site"""

    member_path = 'shipping_methods/%s'
    collection_path = 'shipping_methods'

    nodename = 'shipping_method'

    attributes = (
      'code',
      'name',
      'accounting_code',
      'tax_code',
      'created_at',
      'updated_at',
    )

class Subscription(Resource):

    """A customer account's subscription to your service."""

    member_path = 'subscriptions/%s'
    collection_path = 'subscriptions'

    nodename = 'subscription'

    attributes = (
        'uuid',
        'state',
        'plan_code',
        'coupon_code',
        'coupon_codes',
        'quantity',
        'activated_at',
        'updated_at',
        'canceled_at',
        'starts_at',
        'expires_at',
        'current_period_started_at',
        'current_period_ends_at',
        'trial_started_at',
        'trial_ends_at',
        'unit_amount_in_cents',
        'tax_in_cents',
        'tax_type',
        'tax_rate',
        'total_billing_cycles',
        'remaining_billing_cycles',
        'timeframe',
        'currency',
        'subscription_add_ons',
        'account',
        'pending_subscription',
        'net_terms',
        'collection_method',
        'po_number',
        'first_renewal_date',
        'bulk',
        'terms_and_conditions',
        'customer_notes',
        'vat_reverse_charge_notes',
        'bank_account_authorized_at',
        'redemptions',
        'revenue_schedule_type',
        'gift_card',
        'shipping_address',
        'shipping_address_id',
        'shipping_method_code',
        'shipping_amount_in_cents',
        'started_with_gift',
        'converted_at',
        'no_billing_info_reason',
        'imported_trial',
        'remaining_pause_cycles',
        'paused_at',
        'auto_renew',
        'renewal_billing_cycles',
        'first_billing_date',
        'first_bill_date',
        'next_bill_date',
        'current_term_started_at',
        'current_term_ends_at',
        'custom_fields',
        'gateway_code',
        'transaction_type',
        'billing_info',
    )
    sensitive_attributes = ('number', 'verification_value', 'bulk')

    def preview(self):
        if hasattr(self, '_url'):
            url = self._url + '/preview'
            return self.post(url)
        else:
            url = urljoin(recurly.base_uri(), self.collection_path + '/preview')
            return self.post(url)

    def update_notes(self, **kwargs):
        """
        Updates the notes on the subscription without generating a change
        This endpoint also allows you to update custom fields:

            `
                sub.custom_fields[0].value = 'A new value'
                sub.update_notes()
            `
        """
        for key, val in iteritems(kwargs):
            setattr(self, key, val)
        url = urljoin(self._url, '/notes')
        self.put(url)

    def pause(self, remaining_pause_cycles):
        """Pause a subscription"""
        url = urljoin(self._url, '/pause')
        elem = ElementTreeBuilder.Element(self.nodename)
        elem.append(Resource.element_for_value('remaining_pause_cycles',
                                               remaining_pause_cycles))
        body = ElementTree.tostring(elem, encoding='UTF-8')

        response = self.http_request(url, 'PUT', body, { 'Content-Type':
            'application/xml; charset=utf-8' })

        if response.status not in (200, 201, 204):
            self.raise_http_error(response)

        self.update_from_element(ElementTree.fromstring(response.read()))

    def resume(self):
        """Resume a subscription"""
        url = urljoin(self._url, '/resume')
        self.put(url)

    def _update(self):
        if not hasattr(self, 'timeframe'):
            self.timeframe = 'now'
        return super(Subscription, self)._update()

    def __getpath__(self, name):
        if name == 'plan_code':
            return 'plan/plan_code'
        else:
            return name

    def create_usage(self, sub_add_on, usage):
        """Record the usage on the given subscription add on and update the
        usage object with returned xml"""
        url = urljoin(self._url, '/add_ons/%s/usage' % (sub_add_on.add_on_code,))
        return usage.post(url)

class TransactionBillingInfo(recurly.Resource):
    node_name = 'billing_info'
    attributes = (
        'first_name',
        'last_name',
        'address1',
        'address2',
        'city',
        'state',
        'country',
        'zip',
        'phone',
        'vat_number',
        'first_six',
        'last_four',
        'card_type',
        'month',
        'year',
        'transaction_uuid',
    )


class TransactionAccount(recurly.Resource):
    node_name = 'account'
    attributes = (
        'first_name',
        'last_name',
        'company',
        'email',
        'account_code',
    )
    _classes_for_nodename = {'billing_info': TransactionBillingInfo}

class TransactionDetails(recurly.Resource):
    node_name = 'details'
    attributes = ('account')
    _classes_for_nodename = {'account': TransactionAccount}


class TransactionError(recurly.Resource):
    node_name = 'transaction_error'
    attributes = (
        'id',
        'merchant_message',
        'error_caterogy',
        'customer_message',
        'error_code',
        'gateway_error_code',
    )

class TransactionFraudInfo(recurly.Resource):
    node_name = 'fraud'
    attributes = (
        'score',
        'decision',
    )

class Transaction(Resource):

    """An immediate one-time charge made to a customer's account."""

    member_path = 'transactions/%s'
    collection_path = 'transactions'

    nodename = 'transaction'

    attributes = (
        'uuid',
        'action',
        'account',
        'currency',
        'amount_in_cents',
        'tax_in_cents',
        'status',
        'reference',
        'test',
        'voidable',
        'description',
        'refundable',
        'cvv_result',
        'avs_result',
        'avs_result_street',
        'avs_result_postal',
        'created_at',
        'updated_at',
        'details',
        'transaction_error',
        'type',
        'ip_address',
        'tax_exempt',
        'tax_code',
        'accounting_code',
        'fraud',
        'original_transaction',
        'gateway_type',
        'origin',
        'message',
        'approval_code',
    )
    xml_attribute_attributes = ('type',)
    sensitive_attributes = ('number', 'verification_value',)
    _classes_for_nodename = {
        'details': TransactionDetails,
        'fraud': TransactionFraudInfo,
        'transaction_error': TransactionError
    }

    def _handle_refund_accepted(self, response):
        if response.status != 202:
            self.raise_http_error(response)

        self._refund_transaction_url = response.getheader('Location')
        return self

    def get_refund_transaction(self):
        """Retrieve the refund transaction for this transaction, immediately
        after refunding.

        After calling `refund()` to refund a transaction, call this method to
        retrieve the new transaction representing the refund.

        """
        try:
            url = self._refund_transaction_url
        except AttributeError:
            raise ValueError("No refund transaction is available for this transaction")

        resp, elem = self.element_for_url(url)
        value = self.value_for_element(elem)
        return value

    def refund(self, **kwargs):
        """Refund this transaction.

        Calling this method returns the refunded transaction (that is,
        ``self``) if the refund was successful, or raises a `ResponseError` if
        an error occurred requesting the refund. After a successful call to
        `refund()`, to retrieve the new transaction representing the refund,
        use the `get_refund_transaction()` method.

        """
        # Find the URL and method to refund the transaction.
        try:
            selfnode = self._elem
        except AttributeError:
            raise AttributeError('refund')
        url, method = None, None
        for anchor_elem in selfnode.findall('a'):
            if anchor_elem.attrib.get('name') == 'refund':
                url = anchor_elem.attrib['href']
                method = anchor_elem.attrib['method'].upper()
        if url is None or method is None:
            raise AttributeError("refund")  # should do something more specific probably

        actionator = self._make_actionator(url, method, extra_handler=self._handle_refund_accepted)
        return actionator(**kwargs)


Transaction._classes_for_nodename['transaction'] = Transaction


class Plan(Resource):

    """A service level for your service to which a customer account
    can subscribe."""

    member_path = 'plans/%s'
    collection_path = 'plans'

    nodename = 'plan'

    attributes = (
        'plan_code',
        'name',
        'description',
        'success_url',
        'cancel_url',
        'display_donation_amounts',
        'display_quantity',
        'display_phone_number',
        'bypass_hosted_confirmation',
        'unit_name',
        'payment_page_tos_link',
        'plan_interval_length',
        'plan_interval_unit',
        'trial_interval_length',
        'trial_interval_unit',
        'accounting_code',
        'setup_fee_accounting_code',
        'created_at',
        'updated_at',
        'tax_exempt',
        'tax_code',
        'unit_amount_in_cents',
        'setup_fee_in_cents',
        'total_billing_cycles',
        'revenue_schedule_type',
        'setup_fee_revenue_schedule_type',
        'trial_requires_billing_info',
        'auto_renew',
    )

    def get_add_on(self, add_on_code):
        """Return the `AddOn` for this plan with the given add-on code."""
        url = urljoin(self._url, '/add_ons/%s' % (add_on_code,))
        resp, elem = AddOn.element_for_url(url)
        return AddOn.from_element(elem)

    def create_add_on(self, add_on):
        """Make the given `AddOn` available to subscribers on this plan."""
        url = urljoin(self._url, '/add_ons')
        return add_on.post(url)

class Usage(Resource):

    """A recording of usage agains a measured unit"""

    nodename = 'usage'
    collection_path = 'usages'

    attributes = (
        'measured_unit',
        'amount',
        'merchant_tag',
        'recording_timestamp',
        'usage_timestamp',
        'usage_type',
        'unit_amount_in_cents',
        'usage_percentage',
        'billed_at',
        'created_at',
        'updated_at',
    )

class MeasuredUnit(Resource):

    """A unit of measurement for usage based billing"""

    nodename = 'measured_unit'
    member_path = 'measured_units/%s'
    collection_path = 'measured_units'

    attributes = (
        'id',
        'name',
        'display_name',
        'description',
        'created_at',
        'updated_at',
    )

class AddOn(Resource):

    """An additional benefit a customer subscribed to a particular plan
    can also subscribe to."""

    nodename = 'add_on'

    attributes = (
        'add_on_code',
        'name',
        'display_quantity_on_hosted_page',
        'display_quantity',
        'default_quantity',
        'accounting_code',
        'unit_amount_in_cents',
        'measured_unit_id',
        'usage_type',
        'usage_percentage',
        'add_on_type',
        'tax_code',
        'revenue_schedule_type',
        'optional',
        'created_at',
        'updated_at',
    )

class SubscriptionAddOn(Resource):

    """A plan add-on as added to a customer's subscription.

    Use these instead of `AddOn` instances when specifying a
    `Subscription` instance's `subscription_add_ons` attribute.

    """

    nodename = 'subscription_add_on'
    inherits_currency = True

    attributes = (
        'add_on_code',
        'quantity',
        'unit_amount_in_cents',
        'address',
    )

class Note(Resource):

    """A customer account's notes."""

    nodename = 'note'
    collection_path = 'notes'

    attributes = (
        'message',
        'created_at',
    )

    @classmethod
    def from_element(cls, elem):
        new_note = Note()
        for child_el in elem:
            if not child_el.tag:
                continue
            setattr(new_note, child_el.tag, child_el.text)
        return new_note

class CreditPayment(Resource):

    """A payment from credit"""

    nodename = 'credit_payment'
    collection_path = 'credit_payments'
    member_path = 'credit_payments/%s'

    attributes = (
        'uuid',
        'unit_amount_in_cents',
        'currency',
        'action',
        'created_at',
        'updated_at',
        'voided_at',
    )


class ExportDate(Resource):
    nodename = 'export_date'
    collection_path = 'export_dates'

    attributes = (
        'date',
        'export_files'
    )

    def files(self, date):
        """
        Fetch files for a given date.
        :param date: The date to fetch the export files for
        :return: A list of exported files for that given date or an empty list if not file exists for that date
        """
        url = urljoin(recurly.base_uri() + self.collection_path, '/%s/export_files' % date)
        return ExportDateFile.paginated(url)


class ExportDateFile(Resource):
    nodename = 'export_file'
    collection_path = 'export_files'

    attributes = (
        'name',
        'md5sum',
        'expires_at',
        'download_url'
    )

    def download_information(self):
        """
        Download an export file
        :return: The download information of the file that includes the download URL as well as the expiry time of the
            download URL
        """
        _response, element = ExportDateFile.element_for_url(self._url)
        return ExportDateFile.from_element(element)


Resource._learn_nodenames(locals().values())


def objects_for_push_notification(notification):
    """Decode a push notification with the given body XML.

    Returns a dictionary containing the constituent objects of the push
    notification. The kind of push notification is given in the ``"type"``
    member of the returned dictionary.

    """
    notification_el = ElementTree.fromstring(notification)
    objects = {'type': notification_el.tag}
    for child_el in notification_el:
        tag = child_el.tag
        res = Resource.value_for_element(child_el)
        objects[tag] = res
    return objects
